
#include "SeeDOpenGL.h"
#include "SeeDFrustum.h"
#include "seed_math.h"

//#include "mmgr.h"

SeeDFrustum::SeeDFrustum()
 : savedFOV(0.0f), tanFOV(0.0f)
{
    // empty
}


void
SeeDFrustum::Calculate()
{
    gmVector3 axis[3];
	uint i, j;

	//
	// Calculate the Clipping Matrix
	matrix4x4 projection, model, clipping;

	glGetFloatv(GL_PROJECTION_MATRIX, projection);
	glGetFloatv(GL_MODELVIEW_MATRIX, model);

	MatrixMultiply( model, projection, clipping );

	//
	// Calculate Frustum Planes Based on the Clipping Matrix
	for( i=0; i<4; i++ )
    {
		j = i * 4;

		this->frustumPlanes[SFS_RIGHT][i]  = clipping[3 + j] - clipping[j];
		this->frustumPlanes[SFS_LEFT][i]   = clipping[3 + j] + clipping[j];

		this->frustumPlanes[SFS_BOTTOM][i] = clipping[3 + j] + clipping[1 + j];
		this->frustumPlanes[SFS_TOP][i]    = clipping[3 + j] - clipping[1 + j];

		this->frustumPlanes[SFS_BACK][i]   = clipping[3 + j] - clipping[2 + j];
		this->frustumPlanes[SFS_FRONT][i]  = clipping[3 + j] + clipping[2 + j];
	}

	PlaneNormalize(this->frustumPlanes[SFS_LEFT]);
	PlaneNormalize(this->frustumPlanes[SFS_RIGHT]);
	PlaneNormalize(this->frustumPlanes[SFS_TOP]);
	PlaneNormalize(this->frustumPlanes[SFS_BOTTOM]);
	PlaneNormalize(this->frustumPlanes[SFS_FRONT]);
	PlaneNormalize(this->frustumPlanes[SFS_BACK]);

	//
	// Calculate the Frustum Sphere

    const gmVector3 viewPos(ogl.GetViewPos());

    ogl.GetAxis(axis);
    const gmType FOV = ogl.GetFOV();

	const gmType viewDistance = gmAbs(DistanceToPoint(this->frustumPlanes[SFS_FRONT], viewPos));

	gmType viewLength = -(this->frustumPlanes[SFS_BACK][3] + this->frustumPlanes[SFS_FRONT][3]);
    if( FOV != this->savedFOV ) {
		this->savedFOV = FOV;
        this->tanFOV   = (gmType)tan( (double)(FOV * 0.5f) );
	}

    const gmType viewHeight = (viewDistance + viewLength) * this->tanFOV;
    const gmType radius = gmSqrt(gmSqr(viewHeight) + gmSqr(viewLength));

	viewLength = (viewLength * 0.5f) + viewDistance;

    this->frustumSphere.assign(viewPos + viewLength*axis[0], radius);

/*
	// calculate the length of the fov triangle
	float fDepth  = v_height / tan_fov;

	// calculate the corner of the screen
	float fCorner = sqrt(fWidth * fWidth + fHeight * fHeight);

	// now calculate the new fov
	float fFov = atan(fCorner / fDepth);

	// apply to the cone
	m_frusCone.Axis() = vLookVector;
	m_frusCone.Vertex() = Position();
	m_frusCone.SetConeAngle(fFov);*/


}


FrustumResult
SeeDFrustum::PointInFrustum(const gmVector3& point ) const
{
	for( int i=SFS_NUM_SIDES; --i>=0; )
    {
        const gmVector4 frustumPlane(this->frustumPlanes[i]);

		// calculate the plane equation
		// check if the point is behind a side of the frustum
		if( DotProduct(frustumPlane, point) + frustumPlane[3] <= 0 )
			return F_OUT;
	}

	// the point was inside the frustum
	return F_IN;
}


FrustumResult
SeeDFrustum::SphereInFrustum(const gmVector3& center, gmType radius) const
{
#if 0
	// check against sphere first
	vec4_t d;
	VectorSubtract( frustum_sphere, center, d );
	d[3] = radius + frustum_sphere[3];

//	if( VectorLength(d) < (d[3] * d[3]) )
	if( DotProduct(d, d) < (d[3] * d[3] * d[3] * d[3]) )
		return false;
#endif

	for( int i=SFS_NUM_SIDES; --i>=0; ) {
        const gmVector4 frustumPlane(this->frustumPlanes[i]);

		const gmType distance = DistanceToPoint(frustumPlane, center);

		// if the center of the sphere is farther away from the plane than the radius
		if( distance < -radius )
			return F_OUT;			// distance was greater than the radius

		if( gmAbs(distance) < radius )
			return F_INTERSECT;
	}

	return F_IN;
}


FrustumResult
SeeDFrustum::CubeInFrustum(const gmVector3& c, gmType s) const
{
	for( int i=SFS_NUM_SIDES; --i>=0; ) {
        const gmVector4 f(this->frustumPlanes[i]);

		// check all eight corners of the cube

		if( f[0] * (c[0] - s) + f[1] * (c[1] - s) + f[2] * (c[2] - s) + f[3] > 0 )
			continue;
		if(	f[0] * (c[0] + s) + f[1] * (c[1] - s) + f[2] * (c[2] - s) + f[3] > 0 )
			continue;
		if( f[0] * (c[0] - s) + f[1] * (c[1] + s) + f[2] * (c[2] - s) + f[3] > 0 )
			continue;
		if( f[0] * (c[0] + s) + f[1] * (c[1] + s) + f[2] * (c[2] - s) + f[3] > 0 )
			continue;
		if( f[0] * (c[0] - s) + f[1] * (c[1] - s) + f[2] * (c[2] + s) + f[3] > 0 )
			continue;
		if( f[0] * (c[0] + s) + f[1] * (c[1] - s) + f[2] * (c[2] + s) + f[3] > 0 )
			continue;
		if( f[0] * (c[0] - s) + f[1] * (c[1] + s) + f[2] * (c[2] + s) + f[3] > 0 )
			continue;
		if( f[0] * (c[0] + s) + f[1] * (c[1] + s) + f[2] * (c[2] + s) + f[3] > 0 )
			continue;

		// not in frustum
		return F_OUT;
	}

	return F_IN;
}

FrustumResult
SeeDFrustum::RectInFrustum(const gmVector3& center, const gmVector3& mins, const gmVector3& maxs) const
{
    const gmVector3 amins(center + mins);
    const gmVector3 amaxs(center + maxs);

    gmVector3 box[8] = {
	    gmVector3(amins[0], amaxs[1], amaxs[2]),
	    gmVector3(amaxs[0], amaxs[1], amaxs[2]),
	    gmVector3(amins[0], amaxs[1], amins[2]),
	    gmVector3(amaxs[0], amaxs[1], amins[2]),
	    gmVector3(amins[0], amins[1], amaxs[2]),
	    gmVector3(amaxs[0], amins[1], amaxs[2]),
	    gmVector3(amins[0], amins[1], amins[2]),
	    gmVector3(amaxs[0], amins[1], amins[2])
    };

	return this->RectInFrustum( box );
}

FrustumResult
SeeDFrustum::RectInFrustum(const gmVector3* box) const
{
	int tot_in = 0;

	for( int i=SFS_NUM_SIDES; --i>=0; )
    {
		const gmVector4 frustumPlane(this->frustumPlanes[i]);
		int in_count=8;
		int ptin = 1;

		for( int j=8; --j>=0; )
        {
			if( DistanceToPoint(frustumPlane, box[j]) < 0.0f ) {
				ptin = 0;
				--in_count;
			}

		}
		if( !in_count )
			return F_OUT;

		tot_in += ptin;
	}

	if( tot_in == SFS_NUM_SIDES )
		return F_IN;

	return F_INTERSECT;
}


//frustum_result seed_frustum::PlaneInFrustum( const vec4_t plane ) {
//	return (DotProduct(frust[SFS_FRONT], plane) > 0);
//}


//void seed_frustum::

