
#include <vector>

#include "SeeDFrustum.h"
#include "SeeDKDTree.h"

using std::vector;


KDTreeNode::KDTreeNode()
 : splitPlane(KDNP_X), splitScale(0.0f)
{
    // empty
}

uint
KDTreeNode::GetPointDir(const gmVector3& point) const
{
    const gmType diff = point[this->splitPlane]
                        - this->center[this->splitPlane];

	if( diff >= this->splitScale )
		return KDN_FIRST;
	else
		return KDN_SECOND;
}


uint
KDTreeNode::GetTriDir(const gmVector3& p1, const gmVector3& p2, const gmVector3& p3) const
{
    gmType  c = this->center[this->splitPlane];
	gmType  d1 = p1[this->splitPlane] - c,
	        d2 = p2[this->splitPlane] - c,
	        d3 = p3[this->splitPlane] - c;
	gmType  diff = (d1 + d2 + d3) / 3.0f;

    if( diff >= this->splitScale )
		return KDN_FIRST;
	else
		return KDN_SECOND;
}


void
SeeDKDTree::SplitNode(Node* node, TriangleList& triangleList)
{
	vector<TriangleList>  node_lists(KDN_NUM_NODES, TriangleList());
	TriangleList::iterator pTri, dTri;
	uint i;
	uint num=0;

	node->CalcBox();

    if( node->numTriangles <= 256 ) {
        if( node->GenLeaf(triangleList, this->triangleIndex, tris) ) {
            ++this->numLeaves;
            ++this->numNodes;
		}

		return;
	}

	gmType        splitScale = 0.0f;
    kd_node_plane plane = node->splitPlane;

	// calculate the splitting plane
	for( pTri=triangleList.begin(); pTri!=triangleList.end(); ++pTri )
    {
		const gmType v1 = verts[ pTri->tri[0] ].point[plane] - node->center[plane];
		const gmType v2 = verts[ pTri->tri[1] ].point[plane] - node->center[plane];
		const gmType v3 = verts[ pTri->tri[2] ].point[plane] - node->center[plane];

		splitScale += ( (v1 + v2 + v3) / 3.0f );
	}

    node->splitScale = ( splitScale / (float)(node->numTriangles) );
	gmType offset = node->center[ plane ] + node->splitScale;

	kd_node_plane new_plane;

	switch(plane) {
	case KDNP_X:
		new_plane = KDNP_Y;
		break;
	case KDNP_Y:
		new_plane = KDNP_Z;
		break;
	case KDNP_Z:
		new_plane = KDNP_X;
		break;
	};

	gmVector3 _mins[KDN_NUM_NODES];
	gmVector3 _maxs[KDN_NUM_NODES];

	uint node_nums[KDN_NUM_NODES]={0};

	// step through each triangle and determine where each one goes
	for( pTri=triangleList.begin(); pTri!=triangleList.end(); )
    {
        const gmVector3 v1(verts[ pTri->tri[0] ].point);
        const gmVector3 v2(verts[ pTri->tri[1] ].point);
        const gmVector3 v3(verts[ pTri->tri[2] ].point);

		// get the location of each of the triangle's nodes
		const uint dir = node->GetTriDir( v1, v2, v3 );

		// keep track of where this triangle was
		dTri = pTri++;
		node_lists[dir].splice( node_lists[dir].end(), triangleList, dTri );
		node_nums[dir]++;
		++num;

        AddToBounds(v1, _mins[dir], _maxs[dir]);
        AddToBounds(v2, _mins[dir], _maxs[dir]);
        AddToBounds(v3, _mins[dir], _maxs[dir]);
	}

	for( i=0; i<KDN_NUM_NODES; i++ ) {
		if( node_nums[i] == 0 ) {
			node->nodes[i] = NULL;
			continue;
		}

		Node* new_node = (node->nodes[i] = new Node);

		new_node->splitPlane = new_plane;
        new_node->numTriangles = node_nums[i];

#if 0
        gmVector3 amax(node->maxs + node->center);
        gmVector3 amin(node->mins + node->center);

		VectorCopy(node->center, new_node->center);
        new_node->center = node->center;

		switch( i ) {
		case KDN_FIRST:
			// amax
			// amin--
			amin[plane] = node->center[plane] + node->split_scale;
			new_node->center[plane] += (node->maxs[plane] - node->split_scale) * 0.5f + node->split_scale;
			break;

		case KDN_SECOND:
			// amax--
			amax[plane] = node->center[plane] + node->split_scale;
			new_node->center[plane] -= (node->maxs[plane] + node->split_scale) * 0.5f + node->split_scale;
			// amin
			break;

		default:
			new_node->center[0] = 0;
			break;
		};

        new_node->maxs = amax - new_node->center;
        new_node->mins = amin - new_node->center;
#else
//		VectorCopy(min[i], new_node->mins);
//		VectorCopy(max[i], new_node->maxs);
        new_node->maxs = (_maxs[i] - _mins[i]) * 0.5f;
        new_node->center = _mins[i] + new_node->maxs;
        new_node->mins = _mins[i] - new_node->center;
#endif

		this->SplitNode(new_node, node_lists[i]);
	}
}

uint
SeeDKDTree::DrawNode(Node* node, bool checkChildren) const
{
	uint num = 0;

	if( !node )
		return 0;

	if( checkChildren )
    {
		if( frustum.PointInFrustum(node->center) != F_IN )
        {
#if 0
			const gmType m1 = dot(node->maxs, node->maxs);
			const gmType m2 = dot(node->mins, node->mins);
			const gmType lenSquared = (m1 > m2) ? m1 : m2;
#else
			const gmType lenSquared = dot(node->maxs, node->maxs);
#endif

			switch( frustum.SphereInFrustum(node->center, gmSqrt(lenSquared)) ) {
			case F_OUT:
				node->DrawBox();
				return 0;

			case F_IN:
				checkChildren = false;
				break;

			case F_INTERSECT:
				switch( frustum.RectInFrustum(node->box) ) {
				case F_OUT:
//					node->DrawBox();
					return 0;

				case F_IN:
					checkChildren = false;
					break;

				case F_INTERSECT:
					break;
				};
				break;
			};
		}
	}

    if( node->isLeaf ) {
//		visible_nodes.push_back(node);

		num = node->Draw();
//		if( num ) node->DrawBox();

/*		if( node->num_tris > 0 ) {
			glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB, node->buffer);
			glDrawElements(GL_TRIANGLES, node->num_tris*3, GL_UNSIGNED_INT, NULL);

			num = 1;
		}
*/
	} else {
        switch( node->GetPointDir(this->viewpos) ) {
		case KDN_FIRST:
			num += DrawNode(node->nodes[KDN_FIRST], checkChildren);
			num += DrawNode(node->nodes[KDN_SECOND], checkChildren);
			break;

		case KDN_SECOND:
			num += DrawNode(node->nodes[KDN_SECOND], checkChildren);
			num += DrawNode(node->nodes[KDN_FIRST], checkChildren);
			break;
		};
	}

	return num;
}

