
#include "SeeDOpenGL.h"
#include "SeeDWindow.h"
#include "LogManager.h"
#include "seed_math.h"

bool
SeeDOpenGL::CreateWin(WindowInfo& windowInfo, float fov)
{
	if( !SeeDWindow::CreateWin(windowInfo) )
		return false;

	this->render_data.fov_y = fov;

	if( !this->Initialize() )
		return false;

	this->config.vendor_string     = glGetString(GL_VENDOR);
	this->config.renderer_string   = glGetString(GL_RENDERER);
	this->config.version_string    = glGetString(GL_VERSION);
	this->config.extensions_string = glGetString(GL_EXTENSIONS);

	LogMessage("System Information:");
	LogMessage("  Vendor:     %s",	config.vendor_string);
	LogMessage("  Renderer:   %s",	config.renderer_string);
	LogMessage("  Version:    %s",	config.version_string);
	LogMessage("  Extensions: %s",	config.extensions_string);

	this->config.version = this->config.version_string[2] - '0';

    GLenum error = glewInit();
    if( error != GLEW_OK )
        return false;

	return true;
}


bool
SeeDOpenGL::Initialize()
{
	static PIXELFORMATDESCRIPTOR pfd = {	// Get the pixel format's description
		sizeof(PIXELFORMATDESCRIPTOR),		// Get the size of the structure
		1,									// Version number
		PFD_DRAW_TO_WINDOW |				// Format must support Windows
		PFD_SUPPORT_OPENGL |				// Format must support OpenGL
		PFD_DOUBLEBUFFER,					// Must support Double Buffering
		PFD_TYPE_RGBA,						// Request a RGBA (red,green,blue,alpha) format
		info.bpp,							// Select the bits per pixel
		0, 0, 0, 0, 0, 0,					// Color bits ignored
		0,									// No alpha buffer
		0,									// shift bit ignored
		0,									// No accumulation buffer (advanced)
		0, 0, 0, 0,							// Accumulation bits ignored
		16,									// 16 bit Z-Buffer (Depth Buffer)  
		0,									// No Stencil Buffer (advanced)
		0,									// No Auxiliary Buffer (advanced)
		PFD_MAIN_PLANE,						// The main drawing layer
		0,									// Reserved area
		0, 0, 0								// Layer masks ignored
	};

	if( !(hDC=GetDC(hWnd)) ) {
		LogError("Could Not Create a Device Context");

		return false;
	}

	LogMessage("...Device Context Created");

	GLuint	pixelFormat;

	if( !(pixelFormat = ChoosePixelFormat(hDC, &pfd)) ) {
		LogError("Could Not Find a Suitable Pixel Format");

		return false;
	}

	if( !SetPixelFormat(hDC, pixelFormat, &pfd) ) {
		LogError("Could Not Set Pixel Format");

		return false;
	}

	LogMessage("...Pixel Format Found and Set");

	if( !(hRC=wglCreateContext(hDC)) ) {
		LogError("Could Not Create OpenGL Rendering Context");

		return false;
	}

	if( !wglMakeCurrent(hDC, hRC) ) {
		LogError("Could Not Activate an OpenGL Rendering Context");

		return false;
	}

	LogMessage("...OpenGL Rendering Context Created and Activated");

	RECT rec;
	GetClientRect(hWnd, &rec);
	info.width = rec.right - rec.left;
	info.height = rec.bottom - rec.top;

	Resize( info.width, info.height );

	glShadeModel(GL_SMOOTH);
	glClearColor( 0.5f, 0.5f, 0.5f, 0.5f );
	glClearDepth( 1.0f );
	glEnable(GL_DEPTH_TEST);

	glEnable(GL_CULL_FACE);
//	glCullFace(GL_FRONT);

//	glDepthFunc(GL_LEQUAL);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);


#if 1
	float light_ambient[]	= { 1.0f, 1.0f, 1.0f, 1.0f };
	float light_diffuse[]	= { 1.0f, 1.0f, 1.0f, 1.0f };
//	float light_specular[]	= { 1.0f, 1.0f, 1.0f, 0.0f };
	float light_position[]	= {10.0f,10.0f,500.0f, 0.0f };

	float mat_ambient[]		= { 1.0f, 1.0f, 1.0f, 1.0f };
	float mat_diffuse[]		= { 1.0f, 1.0f, 1.0f, 1.0f };
//	float mat_specular[]	= { 0.5f, 0.5f, 0.5f, 0.0f };
//	float mat_emission[]	= { 0.0f, 0.0f, 0.0f, 0.0f };

//	float shiness[]			= { 50.0f };

	glLightfv(GL_LIGHT0, GL_AMBIENT,	light_ambient);
	glLightfv(GL_LIGHT0, GL_DIFFUSE,	light_diffuse);
//	glLightfv(GL_LIGHT0, GL_SPECULAR,	light_specular);
	glLightfv(GL_LIGHT0, GL_POSITION,	light_position);

//	glMaterialfv(GL_FRONT, GL_EMISSION,	mat_emission );
	glMaterialfv(GL_FRONT, GL_AMBIENT,	mat_ambient);
	glMaterialfv(GL_FRONT, GL_DIFFUSE,	mat_diffuse);
//	glMaterialfv(GL_FRONT, GL_SPECULAR,	mat_specular);
//	glMaterialfv(GL_FRONT, GL_SHININESS,shiness);

	glEnable(GL_LIGHT0);
	glEnable(GL_LIGHTING);

#else

	glDisable(GL_LIGHTING);

#endif

	gmVector4 lmodel_ambient(0.2f, 0.2f, 0.2f, 1.0f);
//	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lmodel_ambient);


	return true;
}


bool
SeeDOpenGL::DestroyWin()
{
	LogMessage("Shutting Down OpenGL...");

	if( setup_list )
		glDeleteLists(setup_list, 1);

	if( ortho_list )
		glDeleteLists(ortho_list, 1);

	if( hRC ) {
		if( !wglMakeCurrent(NULL, NULL) )
			LogError("Could Not Release RC and DC");
		else
			LogMessage("...Released RC and DC");

		::ReleaseDC(hWnd, hDC);

		if( !wglDeleteContext(hRC) )
			LogError("Could Not Release Rendering Context");
		else
			LogMessage("...Released Rendering Context");

		hRC = NULL;
	}

	if( info.fullscreen ) {
		ChangeDisplaySettings( NULL, 0 );	// switch back to desktop
		ShowCursor( TRUE );
	}

	return SeeDWindow::DestroyWin();
}


void SeeDOpenGL::Resize( int width, int height ) {
	if( !height )
		height = 1;					// divide by zero is bad

//	video_data.width = render_data.width = width;
//	video_data.height = render_data.height = height;

	info.width = width;
	info.height = height;

	glViewport(0, 0, width, height);// reset the current viewport

	glMatrixMode(GL_PROJECTION);	// select the projection matrix
	glLoadIdentity();				// reset the projection matrix

	gluPerspective(render_data.fov_y, (GLfloat)width/(GLfloat)height, 4, 4096);

	glMatrixMode(GL_MODELVIEW);		// select the modelview matrix
	glLoadIdentity();				// reset the modelview matrix
}

bool
SeeDOpenGL::ToggleFullscreen()
{
//  DestroyGL();
    this->DestroyWin();

    this->info.fullscreen = !this->info.fullscreen;

    if( !this->Create() || !this->Initialize() )
        return false;

    return true;
}


void
SeeDOpenGL::GetInverseModelView(gmType* matrix)
{
    const gmVector3 angles(this->render_data.v_angles);
    const gmVector3 origin(this->render_data.v_origin);

	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glRotatef( 90, 1, 0, 0 );
	glRotatef( -90, 0, 0, 1 );
	glRotatef(angles[2], 1, 0, 0); // r
	glRotatef(angles[0], 0, 1, 0); // p
	glRotatef(angles[1], 0, 0, 1); // y
	glTranslatef(origin[0], origin[1], origin[2]);
	glGetFloatv(GL_MODELVIEW_MATRIX, matrix);
	glPopMatrix();
}


void
SeeDOpenGL::SetView(const gmVector3& pos, const gmVector3& angles)
{
    this->render_data.v_origin = pos;
    this->render_data.v_angles = angles;
    AnglesToAxis(angles, this->render_data.v_axis);
}


const gmVector4
SeeDOpenGL::CalcFrusSphere(float p_near, float p_far) const
{
	float v_len    = p_far - p_near;
	float v_height = 0.0f - (v_len * tan(render_data.fov_y * 0.5f));

	v_len *= 0.5f;
    const gmVector3 d(v_height, v_height, p_near - v_len);

	v_len += p_near;
    return gmVector4(render_data.v_origin + v_len*render_data.v_axis[0], VectorLength(d));
}


void
SeeDOpenGL::BeginFrame()
{
	if( !this->setup_list ) {
		this->setup_list = glGenLists(1);
		glNewList(this->setup_list, GL_COMPILE);

		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();

		glMatrixMode(GL_MODELVIEW);		// select the modelview matrix
		glLoadIdentity();				// reset the modelview matrix

		glRotatef( -90, 1, 0, 0 );		// put Z going up
		glRotatef(  90, 0, 0, 1 );		// put Z going up

		glEnable(GL_LIGHTING);
		glEnable(GL_CULL_FACE);
		glEnable(GL_DEPTH_TEST);

//		glCullFace(GL_FRONT);
//		glFrontFace(GL_CW);
//		glFrontFace(GL_CCW);

		glEndList();
	}

	glCallList(this->setup_list);

	glRotatef(-this->render_data.v_angles[2], 1, 0, 0); // r
	glRotatef(-this->render_data.v_angles[0], 0, 1, 0); // p
	glRotatef(-this->render_data.v_angles[1], 0, 0, 1); // y
	glTranslatef(-this->render_data.v_origin[0],
                 -this->render_data.v_origin[1],
                 -this->render_data.v_origin[2]);
}


void
SeeDOpenGL::EndFrame()
{
	this->BeginOrtho();
	this->EndOrtho();
}


void
SeeDOpenGL::BeginOrtho()
{
	if( !this->ortho_list ) {
		this->ortho_list = glGenLists(1);
		glNewList(this->ortho_list, GL_COMPILE);

		glPushAttrib(GL_ALL_ATTRIB_BITS);

		glMatrixMode(GL_MODELVIEW);
		glPushMatrix();
		glLoadIdentity();

		glMatrixMode(GL_PROJECTION);
		glPushMatrix();
		glLoadIdentity();

		glDisable(GL_CULL_FACE);
		glDisable(GL_LIGHTING);

		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glColor4f(1, 1, 1, 1);

		glEndList();
	}

	glCallList(this->ortho_list);

	glOrtho(0.0f, this->info.width, this->info.height, 0.0f, -1.0f, 1.0f);
	glMatrixMode(GL_MODELVIEW);

//	glDisable(GL_DEPTH_TEST);
//	glDisable(GL_CULL_FACE);
//	glDisable(GL_BLEND);
//	glEnable(GL_ALPHA_TEST);
}


void
SeeDOpenGL::EndOrtho()
{
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();

	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();

	glPopAttrib();

	glFlush();

	SwapBuffers(this->hDC);
}

