
#include "SeeDOpenGL.h"
#include "SeeDFrustum.h"

#include "SeeDQuadTree.h"

#include <vector>
using std::vector;

byte
QuadtreeNode::GetPointDir(const gmVector3& point)
{
    const gmVector3 p(point - this->center);

	if( p[0] >=0 ) {
		// top
		if( p[1] >= 0 )
			return QN_TOP_LEFT;
		else
			return QN_TOP_RIGHT;
	} else {
		// bottom
		if( p[1] >= 0 )
			return QN_BOTTOM_LEFT;
		else
			return QN_BOTTOM_RIGHT;
	}
}

byte
QuadtreeNode::GetTriDir(const gmVector3& p1, const gmVector3& p2, const gmVector3& p3)
{
	const byte d1 = GetPointDir( p1 );
	const byte d2 = GetPointDir( p2 );
	const byte d3 = GetPointDir( p3 );

	return d1;
}


void
SeeDQuadtree::SplitNode(Node* node, TriangleList& triangleList)
{
	if( !node )
		return;

	node->CalcBox();

    if( node->numTriangles <= 128 )
    {
        if( node->GenLeaf(triangleList, this->triangleIndex, this->tris) ) {
            ++this->numLeaves;
            ++this->numNodes;
		}

		return;
	}

    TriangleList nodeLists[QN_NUM_NODES];
	uint  nodeNumbers[QN_NUM_NODES]={0};
	uint  num=0;

	gmVector3 min[QN_NUM_NODES];
	gmVector3 max[QN_NUM_NODES];

	for( TriangleList::iterator pTri=triangleList.begin(); pTri!=triangleList.end(); ++num )
    {
		const gmVector3 vertex1(verts[ pTri->tri[0] ].point);
		const gmVector3 vertex2(verts[ pTri->tri[1] ].point);
		const gmVector3 vertex3(verts[ pTri->tri[2] ].point);

		// get the location of the triangle
		const byte dir = node->GetTriDir(vertex1, vertex2, vertex3);

		nodeLists[dir].splice(nodeLists[dir].end(), triangleList, pTri++);
		nodeNumbers[dir]++;

		AddToBounds(vertex1, min[dir], max[dir]);
		AddToBounds(vertex2, min[dir], max[dir]);
		AddToBounds(vertex3, min[dir], max[dir]);
	}

    const uint len = triangleList.size();

	for( uint dir=0; dir<QN_NUM_NODES; ++dir )
    {
		if( nodeNumbers[dir] <= 0 ) {
			node->nodes[dir] = NULL;
			continue;
		}

		Node* new_node = (node->nodes[dir] = new Node);
        new_node->numTriangles = nodeNumbers[dir];

		gmVector3 centerOffset;

		switch(dir) {
		case QN_TOP_LEFT:
            centerOffset.assign(node->maxs[0], node->maxs[1], 0.0f);
			break;

		case QN_TOP_RIGHT:
			centerOffset.assign(node->maxs[0], node->mins[1], 0.0f);
			break;

		case QN_BOTTOM_LEFT:
			centerOffset.assign(node->mins[0], node->maxs[1], 0.0f);
			break;

		case QN_BOTTOM_RIGHT:
			centerOffset.assign(node->mins[0], node->mins[1], 0.0f);
			break;
		};

        new_node->center = node->center + 0.5f * centerOffset;
        new_node->mins.assign(node->mins[0]*0.5f, node->mins[1]*0.5f, node->mins[2]);
        new_node->maxs.assign(node->maxs[0]*0.5f, node->maxs[1]*0.5f, node->maxs[2]);

		this->SplitNode(new_node, nodeLists[dir]);
	}
}


uint
SeeDQuadtree::DrawNode(Node* node, bool checkChildren) const
{
	if( !node )
		return 0;

	if( checkChildren ) {
		if( frustum.PointInFrustum(node->center) != F_IN ) {
            const gmType lenSquared = dot(node->maxs, node->maxs);

			switch( frustum.SphereInFrustum(node->center, gmSqrt(lenSquared)) ) {
			case F_OUT:
				node->DrawBox();
				return 0;

			case F_IN:
				checkChildren = false;
				break;

			case F_INTERSECT:
				switch( frustum.RectInFrustum(node->box) ) {
				case F_OUT:
					return 0;

				case F_IN:
					checkChildren = false;
					break;

				case F_INTERSECT:
					break;
				};
				break;
			};
		}
	}

	uint num = 0;

	if( node->isLeaf ) {
		num = node->Draw();
	} else {
		switch( node->GetPointDir(this->viewpos) ) {
		case QN_TOP_LEFT:
			num += DrawNode(node->nodes[QN_TOP_LEFT],     checkChildren);
			num += DrawNode(node->nodes[QN_TOP_RIGHT],    checkChildren);
			num += DrawNode(node->nodes[QN_BOTTOM_LEFT],  checkChildren);
			num += DrawNode(node->nodes[QN_BOTTOM_RIGHT], checkChildren);
			break;

		case QN_TOP_RIGHT:
			num += DrawNode(node->nodes[QN_TOP_RIGHT],    checkChildren);
			num += DrawNode(node->nodes[QN_TOP_LEFT],     checkChildren);
			num += DrawNode(node->nodes[QN_BOTTOM_RIGHT], checkChildren);
			num += DrawNode(node->nodes[QN_BOTTOM_LEFT],  checkChildren);
			break;

		case QN_BOTTOM_LEFT:
			num += DrawNode(node->nodes[QN_BOTTOM_LEFT],  checkChildren);
			num += DrawNode(node->nodes[QN_BOTTOM_RIGHT], checkChildren);
			num += DrawNode(node->nodes[QN_TOP_LEFT],     checkChildren);
			num += DrawNode(node->nodes[QN_TOP_RIGHT],    checkChildren);
			break;

		case QN_BOTTOM_RIGHT:
			num += DrawNode(node->nodes[QN_BOTTOM_RIGHT], checkChildren);
			num += DrawNode(node->nodes[QN_BOTTOM_LEFT],  checkChildren);
			num += DrawNode(node->nodes[QN_TOP_RIGHT],    checkChildren);
			num += DrawNode(node->nodes[QN_TOP_LEFT],     checkChildren);
			break;
		};
	}

	return num;
}
