//

#include "SeeDOpenGL.h"
#include "SeeDTexture.h"
#include "seed_utils.h"

//

SeeDTexture::SeeDTexture()
 : data(NULL)
{
    name[0] = '\0';
}

SeeDTexture::SeeDTexture(uint  _textureType,
                         uint  _internalFormat,
                         uint  _width,
                         uint  _height,
                         uint  _dataFormat,
                         byte* _data)
 : textureType(_textureType), internalFormat(_internalFormat),
   width(_width), height(_height), dataFormat(_dataFormat),
   data(_data), bitspp(24)
{
    name[0] = '\0';
}


SeeDTexture::~SeeDTexture()
{
    this->Destroy();
}


bool
SeeDTexture::Create()
{
    glGenTextures(1, &this->index);
    glBindTexture(this->textureType, this->index);
    glTexParameterf(this->textureType, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	if( GLEW_SGIS_generate_mipmap ) {
		glTexParameterf(this->textureType, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
		glTexParameteri(this->textureType, GL_GENERATE_MIPMAP_SGIS, GL_TRUE);
	} else {
		glTexParameterf(this->textureType, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	}

    switch(this->textureType)
    {
    case GL_TEXTURE_2D:
	    glTexImage2D(GL_TEXTURE_2D, 0, this->internalFormat,
                     this->width, this->height, 0,
                     this->dataFormat, GL_UNSIGNED_BYTE, this->data);
        break;
    default:
        return false;
        break;
    }

    return true;
}

bool
SeeDTexture::Destroy()
{
	SAFE_DELETE_ARRAY(this->data);
    glDeleteTextures(1, &this->index);
    return true;
}


void
SeeDTexture::Bind() const
{
	glBindTexture(this->textureType, this->index);
}


void
SeeDTexture::GetInfo(ImageInfo& info) const
{
    info.width  = this->width;
    info.height = this->height;
    info.bitspp = this->bitspp;
    info.data   = this->data;
    info.index  = this->index;
}


SeeDTexture*
CreateTexture(uint  textureType,
              uint  internalFormat,
              uint  width,
              uint  height,
              uint  dataFormat,
              byte* data)
{
    SeeDTexture* newImage = 
        new SeeDTexture(textureType, internalFormat,
                        width, height,
                        dataFormat, data);

    if( !newImage->Create() )
    {
        delete newImage;
        return NULL;
    }

    return newImage;
}

//-- -- --//


ImageLoader::ImageLoader()
{
    // empty
}

ImageLoader::ImageLoader(const string& filename)
{
}

ImageLoader::~ImageLoader()
{
    // empty
}


typedef enum {
	TYPE_NONE,
	TYPE_TARGA,
	TYPE_JPEG,
	TYPE_NUM_TYPES
} ImageType;


ImageType
GetExtType(const string& filename)
{
    const string ext = GetExtension(filename);

    if( ext == ".tga" )
        return TYPE_TARGA;

	if( ext == ".jpg" ||
		ext == ".jpeg" )
		return TYPE_JPEG;

	return TYPE_NONE;
}


SeeDTexture*
ImageLoader::Load(const string& filename) const
{
	ImageType type = GetExtType(filename);

	if( type == TYPE_NONE )
		return NULL;

    ImageLoadingInfo info;
    info.data = NULL;

	bool working  = true;
	bool loaded   = false;
	ImageType loadType = type;

	while( working && !loaded )
    {
        SAFE_DELETE(info.data);

		switch(loadType)
        {
		case TYPE_TARGA:
			if( !LoadTarga(filename, info) ) {
				if( loadType != type )
					working = false;
				else
					loadType = TYPE_JPEG;
			} else {
				loaded = true;
				working = false;
			}
			break;

		case TYPE_JPEG:
			if( !LoadJPEG(filename, info) ) {
				if( loadType != type )
					working = false;
				else
					loadType = TYPE_TARGA;
			} else {
				loaded = true;
				working = false;
			}
			break;

		default:
			working=false;
			break;
		};
	};

	if( !loaded )
    {
        SAFE_DELETE(info.data);
		return NULL;
	}

    return CreateTexture(GL_TEXTURE_2D,
                         info.internalFormat,
                         info.width,
                         info.height,
                         info.dataFormat,
                         info.data);
}


bool
ImageLoader::Unload(SeeDTexture* texture) const
{
    texture->Destroy();
    delete texture;
    return true;
}


//


ImageGenerator::ImageGenerator()
{
    // empty
}


void
ImageGenerator::SetImageSize(uint width, uint height)
{
    this->imageWidth = width;
    this->imageHeight = height;
}





SeeDTexture*
ImageGenerator::CreateStripe(const gmVector3& color1,
                             const gmVector3& color2,
                             gmType stripeWidth)
{
    const uint bpp = 3;
    const uint imageSize = this->imageWidth*this->imageHeight;
    byte* data = new byte[ imageSize*bpp ];

    const uint sWidth =
        (uint)gmFloor(stripeWidth * this->imageWidth);

    uint x, y, row, column;
    ulong pos = 0;

    bool columnFlag = false;
    bool rowFlag = false;
    byte* current = data;

    for( x=0, column=0; x<this->imageWidth; ++x, ++column )
    {
        if( column >= sWidth ) {
            columnFlag = !columnFlag;
            column = 0;
        }

        for( y=0, row=0; y<this->imageHeight; ++y, ++row, current+=bpp )
        {
            if( row >= sWidth ) {
                rowFlag = !rowFlag;
                row = 0;
            }

            if( rowFlag ) {
                VectorCopy(color1, current);
            } else {
                VectorCopy(color2, current);
            }
        }
    }

    return CreateTexture(GL_TEXTURE_2D, GL_RGB,
               this->imageWidth, this->imageHeight,
               GL_RGB, data);
}

SeeDTexture*
ImageGenerator::CreateChecker(const gmVector3& color1,
                              const gmVector3& color2,
                              gmType stripeWidth)
{
    const uint bpp = 3;
    const uint imageSize = this->imageWidth*this->imageHeight;
    byte* data = new byte[ imageSize*bpp ];

    const uint sWidth =
        (uint)gmFloor(stripeWidth * this->imageWidth);

    uint x, y, row, column;
    ulong pos = 0;

    bool columnFlag = false;
    bool rowFlag = false;
    byte* current = data;

    for( x=0, column=0; x<this->imageWidth; ++x, ++column )
    {
        if( column >= sWidth ) {
            columnFlag = !columnFlag;
            column = 0;
        }

        rowFlag = false;

        for( y=0, row=0; y<this->imageHeight; ++y, ++row, current+=bpp )
        {
            if( row >= sWidth ) {
                rowFlag = !rowFlag;
                row = 0;
            }

            if( rowFlag ^ columnFlag ) {
                VectorCopy(color1, current);
            } else {
                VectorCopy(color2, current);
            }
        }
    }

    return CreateTexture(GL_TEXTURE_2D, GL_RGB,
               this->imageWidth, this->imageHeight,
               GL_RGB, data);
}


inline gmType Noise2D(uint x, uint y)
{
    uint n = x + y * 57;
    n = (n << 13) ^ n;
    return (float)(1.0 - ((n*n*15731 + 789221)
                          + 1376312589 & 0x7fffffff)
                          / 1073741724.0);
}

inline gmType SmoothNoise(uint x, uint y)
{
    const gmType corners = ( Noise2D(x-1, y-1)
                           + Noise2D(x+1, y-1)
                           + Noise2D(x-1, y+1)
                           + Noise2D(x+1, y+1) ) / 16.0f;
    const gmType sides = ( Noise2D(x-1, y) + Noise2D(x+1, y)
                         + Noise2D(x, y-1) + Noise2D(x, y+1) ) / 8.0f;
    const gmType center = Noise2D(x, y) / 4.0f;

    return corners + sides + center;
}

inline gmType interpolate(gmType x, gmType y, gmType s)
{
    return ( x*(3*gmPow(1.0-s, 2) - 2*gmPow(1.0-s, 3))
           + y*(3*gmPow(    s, 2) - 2*gmPow(    s, 3)) );
}

inline gmType GetNoise(gmType x, gmType y)
{
    const uint xint = (uint)x;
    const uint yint = (uint)y;

    const gmType xfrac = x - xint;
    const gmType yfrac = y - yint;

    const gmType x0y0 = SmoothNoise(xint+0, yint+0);
    const gmType x1y0 = SmoothNoise(xint+1, yint+0);
    const gmType x0y1 = SmoothNoise(xint+0, yint+1);
    const gmType x1y1 = SmoothNoise(xint+1, yint+1);

    return interpolate( interpolate(x0y0, x1y0, xfrac),
                        interpolate(x0y1, x1y1, xfrac),
                        yfrac );
}

SeeDTexture*
ImageGenerator::CreateNoise() const
{
    const uint bytespp = 3;
    const uint imageSize = this->imageWidth*this->imageHeight;
    byte* data = new byte[ imageSize*bytespp ];

    uint x, y, z;
    byte* current = data;

    const uint thisRandom = (uint)(gmRand() * 5000.0);
    const gmType freq = 0.05f;
    const gmType pers = 0.65f;
    const uint   octaves = 8;

    for( x=0; x<this->imageWidth; ++x )
    {
        for( y=0; y<this->imageHeight; ++y, current+=bytespp )
        {

            gmType val = 0.0;
            gmType amplitude = 1.0;
            gmType frequencyFactor = freq;

            for( z=0; z<octaves; ++z )
            {
                val += GetNoise(y*frequencyFactor,
                                x*frequencyFactor) * amplitude;
                amplitude *= pers;
                frequencyFactor *= 2.0;
            }

            val = val * 0.5f + 0.5f;

            if( val < 0.0 ) val = -val;
            if( val > 1.0 ) val = 1.0;
            const byte fin = (byte)(val * 255.0);

            VectorSet(current, fin, fin, fin);
        }
    }

    return CreateTexture(GL_TEXTURE_2D, GL_RGB,
               this->imageWidth, this->imageHeight,
               GL_RGB, data);
}


//



