
#include "SeeDWindow.h"
#include "LogManager.h"

//#include "mmgr.h"

LRESULT CALLBACK WindowProc(HWND, UINT, WPARAM, LPARAM);

typedef struct VideoMode_s {
	const char* description;
	int         width;
    int         height;
	int         mode;
} VideoMode;

static VideoMode videoModes [] = {
	{ "Mode 0: 320x240",	320,	240,	0  },
	{ "Mode 1: 400x300",	400,	300,	1  },
	{ "Mode 2: 512x384",	512,	384,	2  },
	{ "Mode 3: 640x480",	640,	480,	3  },
	{ "Mode 4: 800x600",	800,	600,	4  },
	{ "Mode 5: 960x720",	960,	720,	5  },
	{ "Mode 6: 1024x768",	1024,	768,	6  },
	{ "Mode 7: 1152x864",	1152,	864,	7  },
	{ "Mode 8: 1280x960",	1280,	960,	8  },
	{ "Mode 9: 1600x1200",	1600,	1200,	9  },
	{ "Mode 10: 2048x1536", 2048,	1536,	10 }
};

#define NUM_MODES	( (sizeof(videoModes)) / (sizeof(videoModes[0])))

int
ScreenMode(int width, int height)
{
	VideoMode *m = &videoModes[0];
	VideoMode *c = &videoModes[1];

	for( int i=0; i<NUM_MODES; i++ )
    {
		int dw1 = width - m[i+0].width;
		int dw2 = width - m[i+1].width;
		int dh1 = height - m[i+0].height;
		int dh2 = height - m[i+1].height;

		if( dw2 > 0 || dh2 > 0 )		// greater than the next dimensions
			continue;

		if( !dw1 || !dh1 )
			return m[i+0].mode;

		if( !dw2 || !dh2 )
			return m[i+1].mode;
	}

	return 1;
}

bool
ScreenMode(int *width, int *height, int mode)
{
	if( mode < 0 || mode > NUM_MODES )
		return false;

	*width = videoModes[mode].width;
	*height = videoModes[mode].height;

	return true;
}

VideoMode*
ScreenMode( unsigned int mode )
{
	if( mode < 0 || mode > NUM_MODES )
		return &videoModes[0];

	return &videoModes[mode];
}

//

bool
SeeDWindow::CreateWin(WindowInfo& windowInfo)
{
    memcpy(&this->info, &windowInfo, sizeof(WindowInfo) );

    if( !this->Create() )
	    return false;

    memcpy(&windowInfo, &this->info, sizeof(WindowInfo));

    return true;
}


bool
SeeDWindow::Create()
{
	LogMessage("Creating Window...");

	this->CheckMode();

	WNDCLASSEX		winclass;
	DWORD			style;
	DWORD			exStyle;

	RECT rect;
	rect.left	= info.x;
	rect.top	= info.y;
	rect.right	= info.width  + info.x;
	rect.bottom = info.height + info.y;

	hInstance				=	GetModuleHandle(NULL);				// Get the window's instance
	winclass.cbSize			=	sizeof(WNDCLASSEX);
	winclass.style			=	CS_HREDRAW | CS_VREDRAW | CS_OWNDC;	// Redraw the window on resize
	winclass.lpfnWndProc	=	WindowProc;							// handle the WINPROC messages
	winclass.cbClsExtra		=	0;									// No extra win data
	winclass.cbWndExtra		=	0;									// No extra win data
	winclass.hInstance		=	hInstance;							// Set the window's instance
	winclass.hbrBackground	=	(HBRUSH)GetStockObject(DKGRAY_BRUSH);// No background required for OpenGL stuff
	winclass.hIcon			=	LoadIcon( NULL, IDI_APPLICATION );	// Load the default icon
	winclass.hIconSm		=	LoadIcon( NULL, IDI_HAND );
	winclass.hCursor		=	LoadCursor( NULL, IDC_CROSS );		// Load the mouse pointer
	winclass.lpszMenuName	=	NULL;								// No menu
	winclass.lpszClassName	=	info.classname;						// Set the class name

	if( !RegisterClassEx(&winclass) ) {
		LogError("Could not register window class");

		return false;	// failed
	}

	LogMessage("...Registered window class");

	if( info.fullscreen ) {
		VideoMode* sm = ScreenMode(info.mode);

		LogMessage("...trying fullscreen mode...");

		if( !ChangeResolution(sm->width, sm->height, info.bpp) ) {
			LogError("   Fullscreen mode not supported! -- Trying normal window...");

			info.fullscreen = false;
		} else {
			info.fullscreen = true;
		}
	}

	if( info.fullscreen ) {
		exStyle = WS_EX_APPWINDOW;
		style	= WS_POPUP;
		ShowCursor(FALSE);
	} else {
		exStyle = WS_EX_APPWINDOW | WS_EX_WINDOWEDGE;
		style	= WS_OVERLAPPEDWINDOW;
	}

	// Adjust the window size
//	AdjustWindowRectEx( (RECT*)&rect, style, FALSE, exStyle );
	AdjustWindowRect( (RECT*)&rect, style, FALSE );

	info.width = rect.right - rect.left;
	info.height = rect.bottom - rect.top;

#if 0
	if( !(hWnd = ::CreateWindowEx(exStyle, info.classname, info.title, style|WS_CLIPSIBLINGS|WS_CLIPCHILDREN,
	                              info.x, info.y, info.width, info.height,
	                              NULL, NULL, hInstance, NULL)) ) {
#else
	if( !(hWnd = ::CreateWindow(info.classname, info.title, style|WS_CLIPSIBLINGS|WS_CLIPCHILDREN,
	                            info.x, info.y, info.width, info.height,
	                            NULL, NULL, hInstance, NULL)) ) {
#endif
		LogError("Could not create window");

		this->DestroyWin();

		return false;
	}

	LogMessage("...Success Creating %s Window", (info.fullscreen ? "Fullscreen" : "Normal") );

	::ShowWindow(hWnd, SW_SHOW);
	::UpdateWindow(hWnd);

	// success
	return true;
}

bool
SeeDWindow::DestroyWin()
{
	LogMessage("Destroying Window...");

	if( hWnd && !DestroyWindow(hWnd) )
		hWnd = NULL;

	LogMessage("...window destroyed");

	if( hInstance && !UnregisterClass(info.classname, hInstance) )
		hInstance = NULL;

	LogMessage("...class unregistered");

	return true;
}

void
SeeDWindow::GetInfo(WindowInfo &win_info )
{
	memcpy( &win_info, &info, sizeof(WindowInfo) );
}


void
SeeDWindow::CheckMode()
{
	VideoMode* sm = ScreenMode(info.mode);
	info.mode   = sm->mode;
	info.width  = sm->width;
	info.height = sm->height;
}


bool
SeeDWindow::ChangeResolution(int width, int height, BYTE bpp)
{
    if( !this->info.fullscreen )
        return false;
    
    DEVMODE screenSettings;
    memset( &screenSettings, 0, sizeof(screenSettings) );
    
    screenSettings.dmSize		= sizeof(screenSettings);
    screenSettings.dmPelsWidth	= info.width	= width;
    screenSettings.dmPelsHeight	= info.height	= height;
    screenSettings.dmBitsPerPel	= info.bpp		= bpp;
    screenSettings.dmFields		= DM_BITSPERPEL | DM_PELSWIDTH | DM_PELSHEIGHT;
    
    if( ChangeDisplaySettings(&screenSettings, CDS_FULLSCREEN) != DISP_CHANGE_SUCCESSFUL )
        return false;
    
    return true;
}

