#include <string.h>
#include <io.h>
#include <stdio.h>
#include <math.h>

#include "seed_md3.h"
#include "seed_core.h"

//
// for reading purposes
//

#define MD3_IDENT			(('3'<<24)+('P'<<16)+('D'<<8)+'I')
#define MD3_VERSION			15

// limits
#define MD3_MAX_LODS		4
#define MD3_MAX_TRIANGLES	8192	// per surface
#define MD3_MAX_VERTS		4096	// per surface
#define MD3_MAX_SHADERS		256		// per surface
#define MD3_MAX_FRAMES		1024	// per model
#define MD3_MAX_SURFACES	32		// per model
#define MD3_MAX_TAGS		16		// per frame

// vertex scales
#define	MD3_XYZ_SCALE		(1.0/64)

/*
** md3Surface_t
**
** CHUNK			SIZE
** header			sizeof( md3Surface_t )
** shaders			sizeof( md3Shader_t ) * numShaders
** triangles[0]		sizeof( md3Triangle_t ) * numTriangles
** st				sizeof( md3St_t ) * numVerts
** XyzNormals		sizeof( md3XyzNormal_t ) * numVerts * numFrames
*/

typedef struct {
	int		ident;				// 

	char	name[64];			// polyset name

	int		flags;
	int		numFrames;			// all surfaces in a model should have the same

	int		numShaders;			// all surfaces in a model should have the same
	int		numVerts;

	int		numTriangles;
	int		ofsTriangles;

	int		ofsShaders;			// offset from start of md3Surface_t
	int		ofsSt;				// texture coords are common for all frames
	int		ofsXyzNormals;		// numVerts * numFrames

	int		ofsEnd;				// next surface follows
} md3Surface_t;

typedef struct {
	int			ident;
	int			version;

	char		name[64];			// model name

	int			flags;

	int			numFrames;
	int			numTags;
	int			numSurfaces;

	int			numSkins;

	int			ofsFrames;			// offset for first frame
	int			ofsTags;			// numFrames * numTags
	int			ofsSurfaces;		// first surface, others follow

	int			ofsEnd;				// end of file
} md3Header_t;


typedef struct {
	short		xyz[3];
	short		normal;
} md3XyzNormal_t;

#define s (360.0f/255.0f)

bool seed_md3::Load( const char* filename ) {
	FILE			*fp;
	md3Header_t		*header;
	frameInfo_t		*pFrames;
	tagInfo_t		*pTags;
	md3Surface_t	*pSurfaces;

	byte			*_buffer;
	void			*buffer;
	long			filesize;
	int				i;

//	printf( "%s : Loading", filename );

	// todo: eliminate existing data
	loaded = false;

	// try to open the file
	fp = fopen(filename, "rb");
	if( !fp ) {
//		printf( " Failed to load\n" );
		return false;
	}

	filesize = filelength( fileno( fp ) );
	_buffer = new byte[filesize];//malloc(filesize);//new char[filesize];
	if( !_buffer ) {
//		printf( " Failed to allocate memory\n" );
		return false;
	}
	fread( _buffer, sizeof(char), filesize, fp );
	fclose( fp );
//
	buffer = (char *)_buffer;
	header = (md3Header_t *)_buffer;

	if( header->ident != MD3_IDENT ) {
		printf( " incorrect ident\n" );
		delete _buffer;
		return false;
	}

	// eww?
	if( header->numFrames < 1 ) {
		delete _buffer;
		return false;
	}

	numFrames		= header->numFrames;
	numSurfaces		= header->numSurfaces;
	numTags			= header->numTags;
	numTex			= header->numSkins;

	if( numFrames	>	MD3_MAX_FRAMES		||
		numSurfaces	>	MD3_MAX_SURFACES	||
		numTags		>	MD3_MAX_TAGS		||
		numTex		>	MD3_MAX_SHADERS		)
	{
		delete _buffer;
		return false;
	}

	surfaces	= new surface_t[numSurfaces];
	tags		= new tagInfo_t[numFrames * numTags];
	frameInfo	= new frameInfo_t[numFrames];

	pFrames		= (frameInfo_t *)( ((char *)buffer) + header->ofsFrames );
	pTags		= (tagInfo_t *)( ((char *)buffer) + header->ofsTags );
	pSurfaces	= (md3Surface_t *)( ((char *)buffer) + header->ofsSurfaces);

	// frames
	memcpy( frameInfo, pFrames, sizeof(frameInfo_t) * numFrames );

	// tags
	memcpy( tags, pTags, sizeof(tagInfo_t) * numTags * numFrames );

	// surfaces
	for( i=0; i<numSurfaces; i++ ) {
		int j;

		md3XyzNormal_t	*pNormals	= (md3XyzNormal_t *)( ((char *)pSurfaces) + pSurfaces->ofsXyzNormals );
		texture_t		*pShader	= (texture_t *)( ((char *)pSurfaces) + pSurfaces->ofsShaders );
		tri_t			*pTriangles	= (tri_t *)( ((char *)pSurfaces) + pSurfaces->ofsTriangles );
		st_t			*pSt		= (st_t *)( ((char *)pSurfaces) + pSurfaces->ofsSt );

		if( pSurfaces->numVerts		> MD3_MAX_VERTS		||
			pSurfaces->numTriangles > MD3_MAX_TRIANGLES ||
			pSurfaces->numShaders	> MD3_MAX_SHADERS	||
			pSurfaces->numFrames	> numFrames			)
		{
			delete _buffer;
			return false;
		}

		surfaces[i].frames				= new frame_t[pSurfaces->numFrames];
		surfaces[i].orderedTriangles	= new tri_t[pSurfaces->numTriangles];
		surfaces[i].shaders				= new texture_t[pSurfaces->numShaders];
		surfaces[i].texCoords			= new st_t[pSurfaces->numVerts];

		surfaces[i].numVerts	= pSurfaces->numVerts;
		surfaces[i].numTris		= pSurfaces->numTriangles;
		surfaces[i].numTex		= pSurfaces->numShaders;

		// Shaders
		if( pSurfaces->numShaders > 0 ) {
			memcpy( surfaces[i].shaders, pShader, sizeof(texture_t) * pSurfaces->numShaders );
		}

		// Triangles
		if( pSurfaces->numTriangles > 0 ) {
			memcpy( surfaces[i].orderedTriangles, pTriangles, sizeof(tri_t) * pSurfaces->numTriangles );
		}

		if( pSurfaces->numVerts > 0 ) {
			float	*readVert;

			// Texture Coords
			memcpy( surfaces[i].texCoords, pSt, sizeof(st_t) * pSurfaces->numVerts );

			// xyz normals
			for( j=0; j<numFrames; j++ )
			{
				readVert = new float[pSurfaces->numVerts*6];	// room for xyz and normal

				for( int l=0; l<pSurfaces->numVerts; l++, pNormals++ ) {
					char	a, b;
					float	tmp;
					int		num;

					num = l * 6;

					// read in verts
					readVert[num+0] = pNormals->xyz[0] * MD3_XYZ_SCALE;
					readVert[num+1] = pNormals->xyz[1] * MD3_XYZ_SCALE;
					readVert[num+2] = pNormals->xyz[2] * MD3_XYZ_SCALE;

					// calc normals
					a = ((char *)&pNormals->normal)[0];
					b = ((char *)&pNormals->normal)[1];

					tmp = DEG2RAD(a*s);

					readVert[num+3] = cos( tmp );//norm[0];
					readVert[num+4] = sin( tmp );//norm[1];
					readVert[num+5] = cos( DEG2RAD(b*s) );//norm[2];
				}

				surfaces[i].frames[j].verts = readVert;
			}
		}
		pSurfaces = (md3Surface_t *)( ((char *)pSurfaces) + pSurfaces->ofsEnd );
	}

	delete _buffer;

	loaded = true;

	return true;
}

#define VERTEX_INDEX(v) ((v)*6)
#define NORMAL_INDEX(v) ((v)*6+3)

void seed_md3::Draw( drawInfo_t & drawInfo ) {
	surface_t	*pSurf;
	float		*verts;

	int			frame = drawInfo.frame;
//	tri_t		vertIndex;
//	int			num;


	if( !surfaces )
		return;

	if( !loaded )
		return;

	if( frame > numFrames )
		frame = 0;

	for( int i=0; i<numSurfaces; i++ ) {
		pSurf = &surfaces[i];

		// draw one surface at a time
		verts = pSurf->frames[frame].verts;

		glEnableClientState(GL_NORMAL_ARRAY);
		glEnableClientState(GL_VERTEX_ARRAY);
		glNormalPointer( GL_FLOAT, 6*sizeof(GLfloat), &verts[3] );
		glVertexPointer( 3, GL_FLOAT, 6*sizeof(GLfloat), verts );

		glDrawElements(GL_TRIANGLES, pSurf->numTris*3, GL_UNSIGNED_INT, pSurf->orderedTriangles );

//		glBegin(GL_TRIANGLES);
		// draw verts
		for( int j=0; j<pSurf->numTris; j++ )
		{
//			vertIndex = &pSurf->orderedTriangles[j];
//			VectorScale( pSurf->orderedTriangles[j], 6, vertIndex );
//			glArrayElement( vertIndex[0] );
//			glArrayElement( vertIndex[1] );
//			glArrayElement( vertIndex[2] );

/*
			glNormal3f( verts[vertIndex[0] * 6 + 3],
						verts[vertIndex[0] * 6 + 4],
						verts[vertIndex[0] * 6 + 5] );
			glVertex3f( verts[vertIndex[0] * 6 + 0],
						verts[vertIndex[0] * 6 + 1],
						verts[vertIndex[0] * 6 + 2] );

			glNormal3f( verts[vertIndex[1] * 6 + 3],
						verts[vertIndex[1] * 6 + 4],
						verts[vertIndex[1] * 6 + 5] );
			glVertex3f( verts[vertIndex[1] * 6 + 0],
						verts[vertIndex[1] * 6 + 1],
						verts[vertIndex[1] * 6 + 2] );

			glNormal3f( verts[vertIndex[2] * 6 + 3],
						verts[vertIndex[2] * 6 + 4],
						verts[vertIndex[2] * 6 + 5] );
			glVertex3f( verts[vertIndex[2] * 6 + 0],
						verts[vertIndex[2] * 6 + 1],
						verts[vertIndex[2] * 6 + 2] );

			glNormal3fv( &verts[vertIndex[0] + 3] );
			glVertex3fv( &verts[vertIndex[0]] );

			glNormal3fv( &verts[vertIndex[1] + 3] );
			glVertex3fv( &verts[vertIndex[1]] );

			glNormal3fv( &verts[vertIndex[2] + 3] );
			glVertex3fv( &verts[vertIndex[2]] );*/

		}
//		glEnd();
	}

//	message = "DRAW";
}


/*bool LoadMd3( char* filename ) {
	FILE			*fp;
	md3Header_t		header;
	md3Frame_t		*pFrames;
	md3Tag_t		*pTags;
	md3Surface_t	*pSurfaces;
	md3SurfaceData_t	*pSurfData;
	md3Data_t		model;

//	md3MeshHeader	meshheader;
	void			*_buffer;
	void			*buffer;
	long			filesize;
	int				i;

	int				meshStart;
	int				curLocation;

	// try to open the file
	if( (fp=fopen(filename, "rb")) == 0 ) {
		S3Dlog.Output( "Failed to load %s\n", filename );
		return false;
	}

	filesize = filelength( fileno( fp ) );
	_buffer = malloc( filelength( fileno( fp ) ) );
	fread( _buffer, filesize, 1, fp );
	fclose( fp );

	buffer = (char *)_buffer;

	// read in the header
	header = *(md3Header_t *)_buffer;

	if( header.ident != MD3_IDENT ) {
		S3Dlog.Output( "incorrect ident for %s\n", filename );
		return false;
	}

	// get a pointer to, and allocate memory for the frames
	pFrames = (md3Frame_t *)( ((char *)buffer) + header.ofsFrames );
	model.frames = new md3Frame_t[header.numFrames];

	// get a pointer to and allocate memory for tag_'s
	pTags = (md3Tag_t *)( ((char *)buffer) + header.ofsTags );
	model.tags = new md3Tag_t[header.numFrames * header.numTags];

	// save all tag_ data
	for( i=0; i<header.numTags; i++, pTags++ ) {
		model.tags[i] = *pTags;
	}

	// get a pointer to and allocate memory for surfaces
	pSurfaces = (md3Surface_t *)( ((char *)buffer) + header.ofsSurfaces);
	model.surfData = new md3SurfaceData_t[header.numSurfaces];

	// save all surface data
	for( i=0; i<header.numSurfaces; i++ ) {
		int j;

		md3Shader_t *pShader = (md3Shader_t *)( ((char *)pSurfaces) + pSurfaces->ofsShaders );
		if( pSurfaces->numShaders > 0 ) {
			for( j=0; j<pSurfaces->numShaders; j++, pShader++ ) {
				pSurfData->shaders[j] = *pShader;
			}
		}

		pSurfaces = (md3Surface_t *)( ((char *)pSurfaces) + pSurfaces->ofsEnd );
	}



	// read in the header
/*	fread( &header, sizeof(md3Header_t), 1, file );

	header->headerID[4] = 0;
	if( strcmp( header->headerID, "IDP3" ) != 0 )
		return false;

//	ioffset+=(sizeof(md3Header));
	
	buffer = new byte[header->fileSize - 108];
	fread( buffer, header->fileSize-108, 1, file );

	meshStart = curLocation = header->meshStart - 108;
	if( model->meshes != NULL )
		delete model->meshes;

	model->meshes = new md3Mesh[header->numMeshes];

	for( int i=0; i<header->numMeshes; i++ ) {
		strncpy( meshheader->headerID, (char *)(&buffer[curLocation]), 4 );
		meshheader->headerID[4] = 0;
		curLocation += 4;
		if( strcmp( meshheader->headerID, "IDP3" )==0 ) {
			strncpy( model->meshes[i].meshname, (char *)(&buffer[curLocation]), 68 );
			curLocation += 68;
		}
		model->meshes[i].numFrames =	*(int *)(&buffer[curLocation]);	curLocation += sizeof(int);
		model->meshes[i].numTextures =	*(int *)(&buffer[curLocation]);	curLocation += sizeof(int);
		model->meshes[i].numVertex =	*(int *)(&buffer[curLocation]);	curLocation += sizeof(int);
		model->meshes[i].numFaces =		*(int *)(&buffer[curLocation]);	curLocation += sizeof(int);
		meshheader->startFaces =		*(int *)(&buffer[curLocation]);	curLocation += sizeof(int);
		meshheader->startTextures =		*(int *)(&buffer[curLocation]);	curLocation += sizeof(int);
		meshheader->startTexVertices =	*(int *)(&buffer[curLocation]);	curLocation += sizeof(int);
		meshheader->startVertices =		*(int *)(&buffer[curLocation]);	curLocation += sizeof(int);
		meshheader->sizeMesh =			*(int *)(&buffer[curLocation]);	curLocation += sizeof(int);

		curLocation = header->meshStart + meshheader[i].startVertices;
//		if( model->meshes[i].frames->vertices != NULL )
//			delete model->meshes[i].frames->vertices;

//		model->meshes[i].frames->vertices = new
	}
/
	return true;
}*/
