
#include "seed_utils.h"

#include "mdx/matcomp.h"
#include "mdx_model.h"

#include "seed_ogl.h"
#include "seed_media.h"

#include "seed_vbomanager.h"


void mdxm_model::Unload() {
	if( surf_data ) {
		for( int i=0; i<num_surfaces; ++i ) {
			SAFE_DELETE_ARRAY(surf_data[i].childIndexes);
		}

		JUST_DELETE_ARRAY(surf_data);
	}

	if( LODs ) {
		for( int i=0; i<num_lods; ++i ) {
			if( LODs[i].surfaces ) {
				for( int j=0; j<num_surfaces; ++j ) {
					SAFE_DELETE_ARRAY(LODs[i].surfaces[j].verts);
					SAFE_DELETE_ARRAY(LODs[i].surfaces[j].blended_verts);
					SAFE_DELETE_ARRAY(LODs[i].surfaces[j].tex_coords);
					SAFE_DELETE_ARRAY(LODs[i].surfaces[j].tris);
					SAFE_DELETE_ARRAY(LODs[i].surfaces[j].bone_refs);
				}

				JUST_DELETE_ARRAY(LODs[i].surfaces);
			}
		}

		JUST_DELETE_ARRAY(LODs);
	}
}




bool mdxm_model::Load( char *filename ) {

	WorkBuffer		_buffer(filename, "rb");
	mdxmHeader_t	*header;
	mdxmSurface_t	*rSurf;

	byte	*buffer;
	int		i, j;

	buffer = (byte*)_buffer.GetBuffer();
	header = (mdxmHeader_t*)buffer;

	if( header->ident != MDXM_IDENT ) {
		printf(" incorrect file format\n");
		return false;
	}

	num_lods = header->numLODs;
	LODs = new mdx_lod_t[ num_lods ];

	num_surfaces = header->numSurfaces;
	surf_data = new mdx_surface_info_t[ num_surfaces ];

	int *offsets = (int*)(buffer + sizeof(mdxmHeader_t));

	buffer = ((byte*)_buffer.GetBuffer()) + sizeof(mdxmHeader_t);

	root_num=-1;

	for( i=0; i<num_surfaces; ++i ) {
		mdxmSurfHierarchy_t *surf_hier = (mdxmSurfHierarchy_t*)(buffer + offsets[i]);
		memcpy( &surf_data[i], surf_hier, sizeof(mdxmSurfHierarchy_t)-sizeof(int) );

		surf_data[i].childIndexes = new int[ surf_hier->numChildren ];
		memcpy( surf_data[i].childIndexes, surf_hier->childIndexes, sizeof(int)*surf_hier->numChildren );

		surf_data[i].shaderIndex = media.LoadTexture( surf_hier->shader );

		if( surf_data[i].parentIndex == -1 ) {
			if( root_num < 0 )
				root_num = i;
			else
				return false;
		}
	}

	int off = header->ofsLODs;

	buffer = ((byte*)_buffer.GetBuffer());


	for( i=0; i<num_lods; ++i ) {
		offsets = (int*)(buffer + off);

		mdx_surface_t *pSurf = LODs[i].surfaces = new mdx_surface_t[ num_surfaces ];

		for( j=0; j<num_surfaces; ++j, ++pSurf ) {
			rSurf = (mdxmSurface_t*)(buffer + off + offsets[j+1] + sizeof(int));

			memset( pSurf, NULL, sizeof(mdx_surface_t) );

			pSurf->index			= rSurf->thisSurfaceIndex;
			pSurf->num_verts		= rSurf->numVerts;
			pSurf->num_tris			= rSurf->numTriangles;
			pSurf->num_bone_refs	= rSurf->numBoneReferences;

			if( pSurf->num_verts > 0 ) {
				pSurf->verts		= new mdxmVertex_t[ pSurf->num_verts ];
				pSurf->blended_verts= new mdxmVertex_t[ pSurf->num_verts ];
				pSurf->tex_coords	= new st_t[ pSurf->num_verts ];

				char *ptr	= (((char*)rSurf) + rSurf->ofsVerts);
				uint num	= sizeof(mdxmVertex_t) * pSurf->num_verts;

				memcpy( pSurf->verts, ptr, num );
				memcpy( pSurf->tex_coords, (ptr + num), sizeof(st_t) * pSurf->num_verts );

//				for( int k=0; k<pSurf->num_verts; ++k ) {
//					VectorScale( pSurf->verts[k].vertCoords, MD3_XYZ_SCALE, pSurf->verts[k].vertCoords );
//				}
			}

			if( pSurf->num_tris > 0 ) {
				pSurf->tris = new tri_t[ pSurf->num_tris ];

				memcpy( pSurf->tris, (((char*)rSurf) + rSurf->ofsTriangles), sizeof(mdxmTriangle_t) * pSurf->num_tris );
			}

			if( pSurf->num_bone_refs > 0 ) {
				pSurf->bone_refs = new int[ pSurf->num_bone_refs ];

				memcpy( pSurf->bone_refs, (((char*)rSurf) + rSurf->ofsBoneReferences), sizeof(int) * pSurf->num_bone_refs );
			}

		}

		off += offsets[0];
	}

	valid = true;

	return true;
}


void Matrix3x4ToMatrix4x4( const matrix3x4 m1, matrix4x4 m2 ) {
	m2[ 0] = m1[0][0];
	m2[ 1] = m1[1][0];
	m2[ 2] = m1[2][0];
	m2[ 3] = 0;
	m2[ 4] = m1[0][1];
	m2[ 5] = m1[1][1];
	m2[ 6] = m1[2][1];
	m2[ 7] = 0;
	m2[ 8] = m1[0][2];
	m2[ 9] = m1[1][2];
	m2[10] = m1[2][2];
	m2[11] = 0;
	m2[12] = m1[0][3];
	m2[13] = m1[1][3];
	m2[14] = m1[2][3];
	m2[15] = 1;
}


void MatrixMultiply( const matrix3x4 in1, const matrix3x4 in2, matrix3x4 out ) {
	out[0][0] = in1[0][0] * in2[0][0] + in1[0][1] * in2[1][0] + in1[0][2] * in2[2][0];
	out[0][1] = in1[0][0] * in2[0][1] + in1[0][1] * in2[1][1] + in1[0][2] * in2[2][1];
	out[0][2] = in1[0][0] * in2[0][2] + in1[0][1] * in2[1][2] + in1[0][2] * in2[2][2];

	out[1][0] = in1[1][0] * in2[0][0] + in1[1][1] * in2[1][0] + in1[1][2] * in2[2][0];
	out[1][1] = in1[1][0] * in2[0][1] + in1[1][1] * in2[1][1] + in1[1][2] * in2[2][1];
	out[1][2] = in1[1][0] * in2[0][2] + in1[1][1] * in2[1][2] + in1[1][2] * in2[2][2];

	out[2][0] = in1[2][0] * in2[0][0] + in1[2][1] * in2[1][0] + in1[2][2] * in2[2][0];
	out[2][1] = in1[2][0] * in2[0][1] + in1[2][1] * in2[1][1] + in1[2][2] * in2[2][1];
	out[2][2] = in1[2][0] * in2[0][2] + in1[2][1] * in2[1][2] + in1[2][2] * in2[2][2];
}

void Matrix3x4Copy( const matrix3x4 in, matrix3x4 out ) {
	out[0][0] = in[0][0];
	out[0][1] = in[0][1];
	out[0][2] = in[0][2];
	out[0][3] = in[0][3];

	out[1][0] = in[1][0];
	out[1][1] = in[1][1];
	out[1][2] = in[1][2];
	out[1][3] = in[1][3];

	out[2][0] = in[2][0];
	out[2][1] = in[2][1];
	out[2][2] = in[2][2];
	out[2][3] = in[2][3];
}


void Matrix3x4ToQuatBone( const matrix3x4 mat, bone_t *bone ) {
	VectorSet(bone->pos, mat[0][3], mat[1][3], mat[2][3] );
	Matrix3x4ToQuat(mat, bone->axis);

//	Matrix3x4ToMatrix4x4( mat, bone->matrix );
//	Matrix3x4Copy(mat, bone->matrix);
}

void QuatBoneToMatrix4x4( const bone_t *bone, matrix4x4 mat ) {
	QuatToMatrix4x4( bone->axis, mat );
	mat[ 3] = bone->pos[0];
	mat[ 7] = bone->pos[1];
	mat[11] = bone->pos[2];

//	memcpy( mat, bone->matrix, sizeof(matrix4x4) );
//	Matrix3x4ToMatrix4x4( bone->matrix, mat );
}

matrix4x4 matrix={0};


bool mdxm_model::DrawSurface( mdx_surface_t *surface ) {
	int		i;
	mdx_surface_info_t *si = &surf_data[ surface->index ];

	// offset this surface
//	glPushMatrix();

	media.GetTexture( si->shaderIndex )->Bind();

//	bone_t *ref_bones[4];

	if( surface->num_bone_refs > 0 ) {
		for( i=0; i<surface->num_bone_refs; ++i ) {
//			ref_bones[i] = _anims->GetChunkBone( surface->bone_refs[i], working_frame );
		}
	}

	mdxmVertex_t *pVert = surface->verts, *bVert = surface->blended_verts;
//	mdxmVertex_t tVert;

#if 1
	for( i=surface->num_verts; --i>=0; ++pVert, ++bVert ) {
		// blend each vert
		int num_weights = G2_GetVertWeights(pVert);

		VectorCopy(pVert->vertCoords, bVert->vertCoords);
		VectorCopy(pVert->normal, bVert->normal);

		if( num_weights > 0 ) {
			bone_t tBone;
//			VectorSet4(tBone.axis, 1.0f, 0.0f, 0.0f, 0.0f);
			VectorClear(tBone.pos);
			

			float	tot_weight = 0.0f;
			int		index;

			for( int j=0; j<num_weights; ++j ) {
				index = G2_GetVertBoneIndex(pVert, j);

				if( index < 0 )
					continue;

//				bone_t *pBone = &_anims->working_bones[ index ];//ref_bones[ G2_GetVertBoneIndex(pVert, j) ];
				bone_t *pBone = _anims->GetChunkBone( surface->bone_refs[index], working_frame );

//				QuatSlerp( pBone->axis, 
				vec_t weight = G2_GetVertBoneWeight(pVert, j, tot_weight, num_weights);// / (float)num_weights;
				vec_t wmin = 1.0f - weight;

				if( i == 0 ) {
					VectorScale(pBone->pos, weight, tBone.pos);
					VectorCopy4(pBone->axis, tBone.axis);
				} else {
//					VectorScale(tBone.pos, wmin, tBone.pos);
					VectorMA(tBone.pos, weight, pBone->pos, tBone.pos);
					QuatSlerp(tBone.axis, pBone->axis, weight, tBone.axis);
//					QuatVec3Mult(tBone.axis, tBone.pos, tBone.pos);
				}
			}

//			tot_weight = 1.0f / tot_weight;
//			VectorScale(tBone.pos, tot_weight, tBone.pos);

#if 1
			QuatVec3Mult(tBone.axis, pVert->vertCoords, bVert->vertCoords);
			VectorAdd(bVert->vertCoords, tBone.pos, bVert->vertCoords);
#else
			VectorAdd(bVert->vertCoords, tBone.pos, bVert->vertCoords);
			QuatVec3Mult(tBone.axis, pVert->vertCoords, bVert->vertCoords);
#endif

//			VectorAdd(pVert->normal, tBone.pos, bVert->normal);
			QuatVec3Mult(tBone.axis, pVert->normal, bVert->normal);
			VectorNormalize( bVert->normal );
		}
	}
#endif

	// draw this surface
//	glNormalPointer(		GL_FLOAT, sizeof(mdxmVertex_t),	surface->verts->normal );
	glTexCoordPointer(	2,	GL_FLOAT, sizeof(st_t),			surface->tex_coords );
	glVertexPointer(	3,	GL_FLOAT, sizeof(mdxmVertex_t),	surface->blended_verts->vertCoords );

	glDrawElements( GL_TRIANGLES, surface->num_tris*3, GL_UNSIGNED_INT, surface->tris );

	// draw children
	if( si->numChildren > 0 ) {
		for( i=0; i<si->numChildren; ++i )
			DrawSurface( &working_surfaces[ si->childIndexes[i] ] );
	}

//	glPopMatrix();

	// success!
	return true;
}


void mdxm_model::Draw( int frame, int lod ) {
	if( !valid )
		return;

	working_frame = frame;

	working_surfaces = LODs[ lod ].surfaces;

	if( !working_surfaces )
		return;

	glMatrixMode(GL_MODELVIEW);
//	glPushMatrix();

//	glLoadIdentity();
	_anims->SetWorkingFrame(frame);

	glUnbindBufferARB(GL_ARRAY_BUFFER_ARB);
	glUnbindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB);

	DrawSurface( &working_surfaces[ root_num ] );

//	glPopMatrix();
}


void mdxm_animations::Unload() {
	int i;

	if( skeletons ) {
		for( i=0; i<num_bones; ++i ) {
			SAFE_DELETE_ARRAY(skeletons[i].children);
		}

		JUST_DELETE_ARRAY(skeletons);
	}

	if( frames ) {
		for( i=0; i<num_frames; ++i ) {
			SAFE_DELETE_ARRAY(frames[i].index);
		}
	}

	SAFE_DELETE_ARRAY(bones);
	SAFE_DELETE_ARRAY(working_bones);
}

bool mdxm_animations::Load( char *filename ) {
	WorkBuffer		_buffer(filename, "rb");

	mdxaHeader_t	*header;
	mdxaSkel_t		*pSkel;
	mdx_skeleton_t	*tSkel;
	byte	*buffer;
	int		i, j;

	buffer = (byte*)_buffer.GetBuffer();

	header = (mdxaHeader_t*)buffer;

	if( header->ident != MDXA_IDENT || header->version != MDXA_VERSION ) {
//		printf("incorrect file format.\n");
		return false;
	}

	num_bones = header->numBones;
	num_frames = header->numFrames;

	skeletons = new mdx_skeleton_t[ num_bones ];
	working_bones = new bone_t[ num_bones ];

	memset( skeletons, NULL, sizeof(mdx_skeleton_t) * num_bones );

	int *offsets = (int*)(buffer + sizeof(mdxaHeader_t));

	// start at the offsets
	buffer = ((byte*)_buffer.GetBuffer()) + sizeof(mdxaHeader_t);

	for( i=0; i<num_bones; ++i ) {
		pSkel = (mdxaSkel_t*)(buffer + offsets[i]);
		tSkel = &skeletons[i];

		tSkel->parent		= pSkel->parent;
		tSkel->flags		= pSkel->flags;
		tSkel->numChildren	= pSkel->numChildren;
		strcpy(tSkel->name, pSkel->name);

		Matrix3x4ToQuatBone( pSkel->BasePoseMat.matrix, &tSkel->BasePoseMat );
		Matrix3x4ToQuatBone( pSkel->BasePoseMatInv.matrix, &tSkel->BasePoseMatInv );

		if( tSkel->parent == -1 )
			root_num = i;

		if( pSkel->numChildren > 0 ) {
			tSkel->children = new int[ pSkel->numChildren ];
			memcpy( tSkel->children, pSkel->children, sizeof(int) * pSkel->numChildren );
		} else
			tSkel->children = NULL;
	}

	buffer = ((byte*)_buffer.GetBuffer()) + header->ofsFrames;

	frames = new mdx_frame_t[ num_frames ];

	int max = 0,
		off = 0;
	int *iptr;

	for( i=0; i<num_frames; ++i ) {
		frames[i].index = new int[ num_bones ];

		for( j=0, iptr=frames[i].index; j<num_bones; ++j, ++iptr, buffer+=3 ) {
			*iptr = ( *((int*)(buffer)) ) & 0x00FFFFFF;

			if( *iptr > max )
				max = *iptr;
		}
	}

	num_quats = max + 1;

	buffer = ((byte*)_buffer.GetBuffer()) + header->ofsCompBonePool;

	bones = new bone_t[ num_quats ];

	mdxaBone_t bone;

	for( i=0; i<num_quats; ++i, buffer+=sizeof(mdxaCompQuatBone_t) ) {
		MC_UnCompressQuat( bone.matrix, buffer );
		Matrix3x4ToQuatBone( bone.matrix, &bones[i] );
	}

//	memcpy( bones, buffer, sizeof(mdxaCompQuatBone_t) * num_quats );

	working_frame = -1;

	//

	return true;
}


bone_t	*mdxm_animations::GetChunkBone( int bone_id, int frame ) {
	return &bones[ frames[ frame ].index[ bone_id ] ];
}


bool mdxm_animations::SetWorkingFrame( int frame, int bone ) {
	if( bone == -1 ) {
		if( working_frame == frame )
			return false;

		working_frame = frame;
		bone = root_num;
//		memcpy( &working_bones[bone], GetChunkBone(bone, frame), sizeof(bone_t) );
		memcpy( &working_bones[bone], &skeletons[bone].BasePoseMat, sizeof(bone_t) );
	} else {
		mdx_skeleton_t *pSkel = &skeletons[ bone ];

//		bone_t *parent = GetChunkBone(skeletons[bone].parent, frame);
		bone_t *parent = &working_bones[ skeletons[bone].parent ];
		bone_t *curr = GetChunkBone(bone, frame);
		bone_t *work = &working_bones[ bone ];

		matrix4x4	mPar, mCurr, mTemp;

		QuatBoneToMatrix4x4( parent, mPar );
		QuatBoneToMatrix4x4( curr, mCurr );
		MatrixMultiply( mPar, mCurr, mTemp );

//		MatrixMultiply( parent->matrix, curr->matrix, work->matrix );


		// parent * bone = new
//		QuatMultiply( curr->axis, parent->axis, work->axis );
//		VectorAdd( parent->pos, curr->pos, work->pos );
//		QuatMultiply( pSkel->BasePoseMat.axis, parent->axis, work->axis );
//		QuatMultiply( curr->axis, work->axis, work->axis );
//		QuatAdd( pSkel->BasePoseMat.axis, parent->axis, work->axis );
//		QuatVec3Mult( work->axis, curr->pos, work->pos );
//		VectorAdd( parent->pos, work->pos, work->pos );
//		VectorAdd( pSkel->BasePoseMat.pos, work->pos, work->pos );

//		QuatMultiply( parent->axis, curr->axis, work->axis );
//		VectorSet( work->pos, mTemp[3], mTemp[7], mTemp[11] );

//		QuatVec3Mult( parent->axis, pSkel->BasePoseMat.pos, work->pos );
//		VectorAdd( parent->pos, work->pos, work->pos );
	}

	for( int i=skeletons[bone].numChildren; --i>=0; ) {
		SetWorkingFrame( frame, skeletons[bone].children[i] );
	}

	return true;
}


void mdxm_animations::Draw( int frame, int bone ) {
	bone_t *parent, *curr;
//	vec3_t p;

	if( bone == -1 ) {
		bone = root_num;

		SetWorkingFrame( frame );

//		VectorCopy( skeletons[bone].BasePoseMat.pos, p );
	} else {

		parent = GetChunkBone( skeletons[bone].parent, frame );
		curr = GetChunkBone( bone, frame );

//		parent = &working_bones[ skeletons[bone].parent ];
//		curr = &working_bones[ bone ];



//		VectorAdd(pos, curr->pos, p);
//		VectorAdd(pos, skeletons[bone].BasePoseMat.pos, p);
//		QuatVec3Mult( curr->axis, p, p );

		glColor3f( 0.0f, 0.0f, 0.0f );

		glBegin(GL_LINES);
//			glVertex3fv( parent->pos );
//			glVertex3fv( curr->pos );

//			glVertex3f( parent->matrix[3], parent->matrix[7], parent->matrix[11] );
//			glVertex3f( curr->matrix[3], curr->matrix[7], curr->matrix[11] );
		glEnd();
	}

	for( int i=0; i<skeletons[bone].numChildren; ++i )
		Draw( frame, skeletons[bone].children[i] );
}


void mdxm_animations::DumpMDXA() {
	WorkFile wf("dump.cpp", "wt");

	FILE *fp = wf.Stream();
	if( !fp )
		return;

/*	for( int i=0; i<10; ++i ) {
		fprintf(fp, "{ %f, %f, %f }, { %f, %f, %f }, { %f, %f, %f }, { %f, %f, %f }\n\n",
			bones[i].matrix[0][0], bones[i].matrix[0][1], bones[i].matrix[0][2], bones[i].matrix[0][3],
			bones[i].matrix[1][0], bones[i].matrix[1][1], bones[i].matrix[1][2], bones[i].matrix[1][3],
			bones[i].matrix[2][0], bones[i].matrix[2][1], bones[i].matrix[2][2], bones[i].matrix[2][3] );
	}
*/
}

