
#include <memory.h>

#include "SeeDOpenGL.h"
#include "SeedTexture.h"
#include "seed_utils.h"

#include <jpeglib.h>


const gmType colorScale = (gmType)1.0 / (gmType)255.0;


typedef struct TGA_header_s {
	byte	idLength;			// 0
	byte	colorMapType;		// 1
	byte	imageType;			// 2
	byte	colorMapSpec[5];	// 3
	word	xOrigin;			// 8
	word	yOrigin;
	word	width;
	word	height;
	byte	bpp;
	byte	imageDesc;
} TGA_header_t;


enum targa_type_t {
	TT_NONE = 0,
	TT_COLOR_MAPPED = 1,
	TT_TRUE_COLOR = 2,
	TT_GREYSCALE = 3,

	// Run Length Encoded
	TT_RLE_COLOR_MAPPED = 9,
	TT_RLE_TRUE_COLOR = 10,
	TT_RLE_GREYSCALE = 11,
};

gmVector3
GetColorFromBuffer(const byte* buffer, uint bytespp)
{
	gmVector3 currentPixel(0.0, 0.0, 0.0);

	switch( bytespp ) {
	default:
	case 4:
		// not handled
		break;

	case 3:
		currentPixel[2] = buffer[0];
		currentPixel[1] = buffer[1];
		currentPixel[0] = buffer[2];
		break;

	case 2:
		currentPixel[1] = buffer[1];
		currentPixel[0] = buffer[0];
		break;

	case 1:
		currentPixel[0] = buffer[0];
		break;
	}

	return currentPixel * colorScale;
}

union wordSwap {
	word w;
	struct {
		byte b1;
		byte b2;
	} b;
};

word swapWord( word in ) {
	wordSwap sIn, sOut;
	sIn.w = in;
	sOut.b.b1 = sIn.b.b2;
	sOut.b.b2 = sIn.b.b1;
	return sOut.w;
}

// buffer contains the read file data
bool
LoadTarga(const string& filename, ImageLoadingInfo& info)
{
	WorkBuffer _buffer(DefaultExtension(filename, ".tga"), "rb");

	byte* buffer = (byte*)_buffer.GetBuffer();
	if( !buffer )
		return false;

	TGA_header_t  header;

	memcpy(&header, buffer, sizeof(header));
#ifdef Darwin
	// swap the bytes
	header.xOrigin = swapWord(header.xOrigin);
	header.yOrigin = swapWord(header.yOrigin);
	header.width   = swapWord(header.width);
	header.height  = swapWord(header.height);
#endif

	if( header.imageType != TT_TRUE_COLOR && header.imageType != TT_GREYSCALE &&
		header.imageType != TT_RLE_TRUE_COLOR )
		return false;

	const uint bytespp   = header.bpp>>3;
	const uint imageSize = header.width * header.height * bytespp;

	info.data   = new byte[ imageSize ];
	info.width  = header.width;
	info.height = header.height;
	info.bitspp = header.bpp;

	byte* current = info.data;
	if( !current )
		return false;

    buffer += sizeof(header);

	if( header.imageType == TT_TRUE_COLOR || header.imageType == TT_GREYSCALE )
    {
		memcpy(current, buffer, sizeof(byte) * imageSize);
	}
    else if( header.imageType == TT_RLE_TRUE_COLOR )
    {
		byte* fromData = buffer;
        byte* data = current;
		uint  fromPos=0, dataPos=0;
		byte  num;
		uint  i, j;
		uint  bytesCopied = 0;

		while( bytesCopied < imageSize )
        {
			num = fromData[fromPos++];

			if( num & 0x80 ) {
				// Run-Length Packet
				num = (num&0x7F) + 1;

				for( i=0; i<num; ++i ) {
					for( j=0; j<bytespp; ++j, ++bytesCopied )
						data[dataPos++] = fromData[fromPos + j];
				}

				fromPos += bytespp;
			} else {
				// RAW packet
				++num;
				for( i=0; i<num; ++i )
					for( j=0; j<bytespp; ++j, ++bytesCopied )
						data[dataPos++] = fromData[fromPos++];
			}
		}
	}

	if( !(GLEW_EXT_bgra) )
    {
        // Swap bytes (r and b)
	    byte temp;
		for( uint i=0; i<imageSize; i+=bytespp ) {
			temp = current[i];
			current[i] = current[i+2];
			current[i+2] = temp;
		}

        info.dataFormat = ( bytespp == 3 ? GL_RGB : GL_RGBA );
	} else
		info.dataFormat = ( bytespp == 3 ? GL_BGR_EXT : GL_BGRA_EXT );

	if( GLEW_ARB_texture_compression ) {
        info.internalFormat = ( bytespp == 3 ? GL_COMPRESSED_RGB_ARB : GL_COMPRESSED_RGBA_ARB );
	} else {
		info.internalFormat = ( bytespp == 3 ? GL_RGB : GL_RGBA );
	}

	return true;
}


bool
LoadJPEG(const string& filename, ImageLoadingInfo& info)
{
	WorkFile workFile(DefaultExtension(filename, ".jpg"), "rb");

	if( !workFile.Stream() )
	    return false;

	jpeg_decompress_struct cinfo;
	jpeg_error_mgr jerr;

	cinfo.err = jpeg_std_error(&jerr);

	jpeg_create_decompress(&cinfo);
	jpeg_stdio_src(&cinfo, workFile.Stream());
	jpeg_read_header(&cinfo, TRUE);

	// set color format
	cinfo.out_color_space = JCS_RGB;
	jpeg_start_decompress(&cinfo);

	const uint rowSpan = cinfo.image_width * cinfo.num_components;
	const uint width   = cinfo.image_width;
	const uint height  = cinfo.image_height;

	byte* data = new byte[ rowSpan * height ];

	byte** rowPtr = new byte*[ height ];

	for( uint i = 0; i < height; ++i )
	    rowPtr[i] = &(data[ i * rowSpan ]);

	int rowsRead = 0;
	while( cinfo.output_scanline < cinfo.output_height )
	    rowsRead += jpeg_read_scanlines( &cinfo, &rowPtr[rowsRead], cinfo.output_height - rowsRead );

	delete [] rowPtr;

	jpeg_finish_decompress(&cinfo);
	jpeg_destroy_decompress(&cinfo);

    info.data           = data;
	info.width          = width;
	info.height         = height;
	info.dataFormat     = GL_RGB;
	info.internalFormat = ( GLEW_ARB_texture_compression ) ? GL_COMPRESSED_RGB_ARB : GL_RGB;

	return true;
}

