#include <string.h>
#include <io.h>
#include <stdio.h>
#include <math.h>

#include "SeeDMD3.h"
#include "SeeDOpenGL.h"
#include "SeeDVBOManager.h"

//#include "mmgr.h"

//
// for reading purposes
//


// limits
#define MD3_MAX_LODS		4
#define MD3_MAX_TRIANGLES	8192	// per surface
#define MD3_MAX_VERTS		4096	// per surface
#define MD3_MAX_SHADERS		256		// per surface
#define MD3_MAX_FRAMES		1024	// per model
#define MD3_MAX_SURFACES	32		// per model
#define MD3_MAX_TAGS		16		// per frame

// vertex scales
#define	MD3_XYZ_SCALE		(1.0/64)

/*
** md3Surface_t
**
** CHUNK			SIZE
** header			sizeof( md3Surface_t )
** shaders			sizeof( md3Shader_t ) * numShaders
** triangles[0]		sizeof( md3Triangle_t ) * numTriangles
** st				sizeof( md3St_t ) * numVerts
** XyzNormals		sizeof( md3XyzNormal_t ) * numVerts * numFrames
*/

typedef struct md3Surface_s {
	int		ident;				// 

	char	name[64];			// polyset name

	int		flags;
	int		numFrames;			// all surfaces in a model should have the same

	int		numShaders;			// all surfaces in a model should have the same
	int		numVerts;

	int		numTriangles;
	int		ofsTriangles;

	int		ofsShaders;			// offset from start of md3Surface_t
	int		ofsSt;				// texture coords are common for all frames
	int		ofsXyzNormals;		// numVerts * numFrames

	int		ofsEnd;				// next surface follows
} md3Surface_t;

typedef struct md3Header_s {
	int			ident;
	int			version;

	char		name[64];			// model name

	int			flags;

	int			numFrames;
	int			numTags;
	int			numSurfaces;

	int			numSkins;

	int			ofsFrames;			// offset for first frame
	int			ofsTags;			// numFrames * numTags
	int			ofsSurfaces;		// first surface, others follow

	int			ofsEnd;				// end of file
} md3Header_t;


typedef struct md3XyzNormal_s {
	short		xyz[3];
	short		normal;
} md3XyzNormal_t;

#define s (360.0f/255.0f)

bool SeeDMD3::Load( void *buffer ) {
	md3Header_t		*header;
	frameInfo_t		*pFrames;
	tagInfo_t		*pTags;
	md3Surface_t	*pSurfaces;
	int				i;

	// todo: eliminate existing data
	loaded = false;

	header = (md3Header_t *)buffer;

	if( header->ident != MD3_IDENT )
		return false;

	// eww?
	if( header->numFrames < 1 )
		return false;

	numFrames		= header->numFrames;
	numSurfaces		= header->numSurfaces;
	numTags			= header->numTags;
	numTex			= header->numSkins;

	if( numFrames	>	MD3_MAX_FRAMES		||
		numSurfaces	>	MD3_MAX_SURFACES	||
		numTags		>	MD3_MAX_TAGS		||
		numTex		>	MD3_MAX_SHADERS		)
	{
		return false;
	}

	surfaces	= new md3_surface[numSurfaces];
	tags		= new tagInfo_t[numFrames * numTags];
	frameInfo	= new frameInfo_t[numFrames];

	pFrames		= (frameInfo_t *)( ((char *)buffer) + header->ofsFrames );
	pTags		= (tagInfo_t *)( ((char *)buffer) + header->ofsTags );
	pSurfaces	= (md3Surface_t *)( ((char *)buffer) + header->ofsSurfaces);

	// frames
	memcpy( frameInfo, pFrames, sizeof(frameInfo_t) * numFrames );

	// tags
	memcpy( tags, pTags, sizeof(tagInfo_t) * numTags * numFrames );

	// surfaces
	for( i=0; i<numSurfaces; i++ ) {
		int j;

		md3XyzNormal_t*  pNormals   = (md3XyzNormal_t*)( ((char *)pSurfaces) + pSurfaces->ofsXyzNormals );
		texture_t*       pShader    = (texture_t*)( ((char *)pSurfaces) + pSurfaces->ofsShaders );
		Triangle*        pTriangles = (Triangle*)( ((char *)pSurfaces) + pSurfaces->ofsTriangles );
		TexCoord*        pSt        = (TexCoord*)( ((char *)pSurfaces) + pSurfaces->ofsSt );

		if( pSurfaces->numVerts		> MD3_MAX_VERTS		||
			pSurfaces->numTriangles > MD3_MAX_TRIANGLES ||
			pSurfaces->numShaders	> MD3_MAX_SHADERS	||
			pSurfaces->numFrames	> numFrames			)
		{
			return false;
		}

		surfaces[i].frames				= new frame_t[pSurfaces->numFrames];
		surfaces[i].orderedTriangles	= new Triangle[pSurfaces->numTriangles];
		surfaces[i].shaders				= new texture_t[pSurfaces->numShaders];
		surfaces[i].texCoords			= new TexCoord[pSurfaces->numVerts];

		surfaces[i].numFrames  = numFrames;
		surfaces[i].numVerts   = pSurfaces->numVerts;
		surfaces[i].numTris    = pSurfaces->numTriangles;
		surfaces[i].numTex     = pSurfaces->numShaders;

		// Shaders
		if( pSurfaces->numShaders > 0 )
			memcpy( surfaces[i].shaders, pShader, sizeof(texture_t) * pSurfaces->numShaders );

		// Triangles
		if( pSurfaces->numTriangles > 0 )
			memcpy( surfaces[i].orderedTriangles, pTriangles, sizeof(Triangle) * pSurfaces->numTriangles );

		if( pSurfaces->numVerts > 0 ) {
			float	*readVert;

			// Texture Coords
			for( j=0; j<pSurfaces->numVerts; j++, pSt++ ) {
				surfaces[i].texCoords[j][0] = pSt[0][0];
				surfaces[i].texCoords[j][1] = 1.0f - pSt[0][1];
			}

//			memcpy( surfaces[i].texCoords, pSt, sizeof(st_t) * pSurfaces->numVerts );

			// xyz normals
			for( j=0; j<numFrames; j++ )
			{
				readVert = new float[pSurfaces->numVerts*6];	// room for xyz and normal

				for( int l=0; l<pSurfaces->numVerts; l++, pNormals++ ) {
					char	a, b;
					float	tmp;
					int		num;

					num = l * 6;

					// read in verts
					readVert[num+0] = pNormals->xyz[0] * MD3_XYZ_SCALE;
					readVert[num+1] = pNormals->xyz[1] * MD3_XYZ_SCALE;
					readVert[num+2] = pNormals->xyz[2] * MD3_XYZ_SCALE;

					// calc normals
					a = ((char *)&pNormals->normal)[0];
					b = ((char *)&pNormals->normal)[1];

					tmp = DEG2RAD(a*s);

					readVert[num+3] = cos( tmp );//norm[0];
					readVert[num+4] = sin( tmp );//norm[1];
					readVert[num+5] = cos( DEG2RAD(b*s) );//norm[2];
				}

				surfaces[i].frames[j].verts = readVert;

				glGenBuffersARB(1, &surfaces[i].vertex_buffer);
				glBindBufferARB(GL_ARRAY_BUFFER_ARB, surfaces[i].vertex_buffer);
				glBufferDataARB(GL_ARRAY_BUFFER_ARB, pSurfaces->numVerts*sizeof(float)*6, readVert, GL_STATIC_DRAW_ARB);
				glUnbindBufferARB(GL_ARRAY_BUFFER_ARB);
			}
		}

		pSurfaces = (md3Surface_t *)( ((char *)pSurfaces) + pSurfaces->ofsEnd );
	}

	loaded = true;


	return true;
}


bool SeeDMD3::Load( const char* filename ) {
	FILE			*fp;
	byte			*_buffer;
	void			*buffer;
	long			filesize;

	// todo: eliminate existing data
	loaded = false;

	// try to open the file
	fp = fopen( filename, "rb" );
	if( !fp )
		return false;

	filesize = filelength( fileno(fp) );
	_buffer = new byte[ filesize ];
	if( !_buffer )
		return false;

	fread( _buffer, sizeof(char), filesize, fp );
	fclose( fp );

	buffer = (char *)_buffer;

	Load( buffer );

	delete [] _buffer;

	return loaded;
}


#define VERTEX_INDEX(v) ((v)*6)
#define NORMAL_INDEX(v) ((v)*6+3)

bool SeeDMD3::Draw( drawInfo_t & drawInfo ) {
	md3_surface	*pSurf;
	float		*verts;

	int			frame = drawInfo.frame;
//	tri_t		vertIndex;
//	int			num;

	glUnbindBufferARB(GL_ARRAY_BUFFER_ARB);
	glUnbindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB);

	if( !surfaces )
		return false;

	if( !loaded )
		return false;

	if( frame < 0 || frame > numFrames )
		frame = 0;

	glColor3f(0.5f,0.5f,1.0f);

	pSurf = surfaces;
	for( int i=0; i<numSurfaces; ++i, ++pSurf ) {
		// draw one surface at a time
		verts = pSurf->frames[frame].verts;

		glEnableClientState(GL_NORMAL_ARRAY);
		glEnableClientState(GL_VERTEX_ARRAY);
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);

		glNormalPointer(     GL_FLOAT, 6*sizeof(GLfloat), &verts[3] );
		glTexCoordPointer(2, GL_FLOAT, sizeof(TexCoord),  pSurf->texCoords );
		glVertexPointer(  3, GL_FLOAT, 6*sizeof(GLfloat), verts );

		glDrawElements(GL_TRIANGLES, pSurf->numTris*3, GL_UNSIGNED_INT, pSurf->orderedTriangles);

#if 0
		// draw verts
		for( int j=0; j<pSurf->numTris; j++ )
		{
//			vertIndex = &pSurf->orderedTriangles[j];
			tri_t vertIndex;
			VectorScale( pSurf->orderedTriangles[j], 6, vertIndex );

			glArrayElement( vertIndex[0] );
			glArrayElement( vertIndex[1] );
			glArrayElement( vertIndex[2] );

#if 0
			glNormal3f( verts[vertIndex[0] * 6 + 3],
						verts[vertIndex[0] * 6 + 4],
						verts[vertIndex[0] * 6 + 5] );
			glVertex3f( verts[vertIndex[0] * 6 + 0],
						verts[vertIndex[0] * 6 + 1],
						verts[vertIndex[0] * 6 + 2] );

			glNormal3f( verts[vertIndex[1] * 6 + 3],
						verts[vertIndex[1] * 6 + 4],
						verts[vertIndex[1] * 6 + 5] );
			glVertex3f( verts[vertIndex[1] * 6 + 0],
						verts[vertIndex[1] * 6 + 1],
						verts[vertIndex[1] * 6 + 2] );

			glNormal3f( verts[vertIndex[2] * 6 + 3],
						verts[vertIndex[2] * 6 + 4],
						verts[vertIndex[2] * 6 + 5] );
			glVertex3f( verts[vertIndex[2] * 6 + 0],
						verts[vertIndex[2] * 6 + 1],
						verts[vertIndex[2] * 6 + 2] );
#endif

#if 0
		glBegin(GL_TRIANGLES);
//			glNormal3fv( &verts[vertIndex[0] + 3] );
			glVertex3fv( &verts[vertIndex[0]] );

//			glNormal3fv( &verts[vertIndex[1] + 3] );
			glVertex3fv( &verts[vertIndex[1]] );

//			glNormal3fv( &verts[vertIndex[2] + 3] );
			glVertex3fv( &verts[vertIndex[2]] );
		glEnd();
#endif

		}
#endif

	}

//	message = "DRAW";

	return true;
}


